(*----------------------------------------------------------------------------

  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.uu.nl

  Last update: june 1999

  Here, a Unity 'machine' is modelled by a finite set of actions. 
  The actions are:
     - atomic
     - assumed terminating 
     - can be non-deterministic 
     - always enabled, even if their effect in some situation is 
       just skip 

  Unlike our previous model, we no longer maintain the set of read and
  write variables explicitly in the model since they can be just as well
  decribed by 'confinement' and 'invisible' properties of the program.

     - a finite set of read variables
     - a finite set of write variables. A write variable is assumed to be 
       readable.

  Unlike the original UNITY model, we do not include an "init" section
  in our model to enable us to analyze the model independent from its
  initial condition. In our model, it is as if the initial condition is
  a parameter of the program which can be setup as situation requires.
  This has advantage that when analyzing the program with different 
  initial condition, some analysis can be shared. 

  This module defines the basic operators to abstract the behavior of a
  UNITY machine. 

     - stable  
     - unless  
     - ensures 
     - leadsto  
     - PAR     
  
   Their definition basically follows Chandy & Misra's original, but we
   also put some additional information, namely some kind of 'typing'
   information. More precisely, those variables now has an extra parameter,
   namely a set of variables V. The behavior they describe is now normalized
   by ignoring constraint on variables outside V. 
 
----------------------------------------------------------------------------*)

open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

val _ = new_theory "UNITY99" ;

(*
load "bossLib" ; 
load "uvarsTheory" ;
load "restr_binderTheory" ;
load "upredTheory" ;
load "urelTheory" ;
load "uwfTheory" ;
load "utdcTheory" ;
load "btacs" ;
*)

open bossLib ;
open uvarsTheory ; 
open restr_binderTheory ;
open upredTheory ;
open urelTheory ;
open uwfTheory ;
open utdcTheory ;
open btacs ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;    


(*-- some type abbreviations --*)

val State      = ty_antiq (Type `:'var->'val`) ;
val Pred       = ty_antiq (Type `:^State->bool`) ;
val Action     = ty_antiq (Type `:^State->^State->bool`) ;
val VARS       = ty_antiq (Type `:'var->bool`) ;
val Umachine   = ty_antiq (Type `:^Action->bool`) ;


(*-- define parameterized everywhere operator --*)

val Pvalid_def = Define
   `Pvalid (V:^VARS) (J:^Pred) (p:^Pred) 
    =
    |== (J IMP (p pj V))` ;


(*-- define Hoare triple --*)

val HOA_def = Define 
   `HOA (a:^Action) (p:^Pred) (q:^Pred) 
    = 
    !s t. p s /\ a s t ==> q t` ;


(*-------------------------------------------------------------------
   some basic laws for Hoare triple
---------------------------------------------------------------------*)

val HOA_AND = store_thm
  ("HOA_AND",
 --`HOA a (p:^Pred) (q AND r) 
    = 
    HOA a p q /\ HOA a p r`--,
    REWRITE_TAC [HOA_def, AND_def] THEN BETA_TAC
    THEN PROVE_TAC[]) ;

val HOA_OR = store_thm
  ("HOA_OR",
 --`HOA a (p OR q) (r:^Pred) 
    = 
    HOA a p r /\ HOA a q r`--,
    REWRITE_TAC [HOA_def, OR_def] THEN BETA_TAC
    THEN PROVE_TAC[]) ;


(*-------------------------------------------------------------------
    PRED_TAC_v3 defs lems  tries to prove a given predicate statement
 
        defs : additional definitions which may need expansion
               before first order prover is called
        lems : additional lemmas to be supplied to the first order
               prover
-------------------------------------------------------------------*)

fun PRED_TAC_v3 defs lems =
    let
    (*-- all definitions used to first rewrite goal to first order
         form --*)
    val all_defs = defs @ 
               [Pvalid_def, pSEQ_DEF, NOT_def, AND_def, OR_def, 
                EQUAL_def, IMP_def, RES_qAND_def, RES_qOR_def,
                TT_def, FF_def, 
                restr_binderTheory.RES_FORALL,
                restr_binderTheory.RES_EXISTS]
    in
    (*-- (1) expand functional equalities using extensionality --*)
    CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
    (*-- (2) expand all definitions --*)
    THEN REWRITE_TAC all_defs THEN BETA_TAC THEN REWRITE_TAC all_defs 
    THEN REPEAT STRIP_TAC
    THEN TRY (PROVE_TAC lems) 
    end ;
    
(*-------------------------------------------------------------------
  An action is called always enabled if its execution in any state 
  always leads to some state. 
--------------------------------------------------------------------*)

val ALWAYS_ENABLED_def = Define
   `ALWAYS_ENABLED (a:^Action) = !s. ?t. a s t` ;

(*-- a program ignores a set of variables V if ...               --*)

val IGNORE_def = Define
   `IGNORE (Pr:^Umachine) J V =
    (!a p. Pr a ==> HOA a (J AND (p pj V)) (p pj V))` ;

(*-- a program can only update variables in V by setting a flag q--*)

val CONDWRITE_def = Define
   `CONDWRITE (Pr:^Umachine) J V p q =
    (!a z. Pr a ==> HOA a (J AND p AND (z pj V)) 
                          ((p AND (z pj V)) OR q))` ;

(*-- you can express IGNORE in terms of CONDWRITE                --*)

val IGNORE_adef1 = store_thm
  ("IGNORE_adef1",
 --`IGNORE (Pr:^Umachine) J V = CONDWRITE Pr J V TT FF`--,

    PRED_TAC_v3 [HOA_def, IGNORE_def, CONDWRITE_def] []) ;

(*-- the new definition of ignore is equal to the old/classic one --*)

val IGNORE_adef2 = store_thm
  ("IGNORE_adef2",
 --`IGNORE (Pr:^Umachine) J V 
    =
    (!a X. Pr a ==> HOA a (J AND (\s. s Pj V = X)) (\s. s Pj V = X))`--,

    REWRITE_TAC [HOA_def, AND_def, IGNORE_def, pj_def]
    THEN BETA_TAC THEN EQ_TAC 
    THENL
    [ REPEAT STRIP_TAC
      THEN XRULE_ASSUM_TAC 
           (fn th=> ISPECL [--`a:^Action`--,
                            --`(\s:^State. s = X)`--] th)
      THEN PROVE_TAC[],
      REPEAT STRIP_TAC
      THEN XRULE_ASSUM_TAC 
           (fn th=> ISPECL [--`a:^Action`--,
                            --`(s:^State) Pj V`--] th)  
      THEN PROVE_TAC[] ]
    ) ;    

(*-- the new definition of ignore is also equal to the old/classic one --*)

val CONDWRITE_adef1 = store_thm
  ("CONDWRITE_adef1",
 --`CONDWRITE (Pr:^Umachine) J V p q 
    =
    (!a X. Pr a ==> HOA a (J AND p AND (\s. s Pj V = X)) 
                          ((p AND (\s. s Pj V = X)) OR q))`--,
   
    REWRITE_TAC [HOA_def, AND_def, OR_def, CONDWRITE_def, pj_def]
    THEN BETA_TAC THEN EQ_TAC 
    THENL
    [ REPEAT STRIP_TAC
      THEN XRULE_ASSUM_TAC 
           (fn th=> ISPECL [--`a:^Action`--,
                            --`(\s:^State. s = X)`--] th)
      THEN PROVE_TAC[],
      REPEAT STRIP_TAC
      THEN XRULE_ASSUM_TAC 
           (fn th=> ISPECL [--`a:^Action`--,
                            --`(s:^State) Pj V`--] th)  
      THEN PROVE_TAC[] ]
    ) ;   

(*-------------------------------------------------------------------

  Now, let us define our behavior operators.

---------------------------------------------------------------------*)

val STABLE_def = Define
   `STABLE (Pr:^Umachine) J   =   !a::Pr. HOA a J J` ;


val UNLESS_def = Define
   `UNLESS (Pr:^Umachine) 
           (V:^VARS) 
           (J:^Pred) 
           (p:^Pred) 
           (q:^Pred) 
    =
    STABLE Pr J /\
    (!a::Pr. HOA a (J AND (p pj V) AND NOT(q pj V)) 
                   ((p pj V) OR (q pj V)))` ;


val ENSURES_def = Define
   `ENSURES (Pr:^Umachine) 
            (V:^VARS) 
            (J:^Pred) 
            (p:^Pred) 
            (q:^Pred) 
    =
    UNLESS Pr V J p q /\ 
    (?a::Pr. HOA a (J AND (p pj V) AND NOT(q pj V)) 
                   (q pj V))` ;

val LEADSTO_def = Define
   `(LEADSTO:^Umachine->^VARS->^Pred->^Pred->^Pred->bool)
     
      Pr V J
      = 
      TDC (ENSURES Pr V J)` ;    


val UNTIL_def = Define
   `UNTIL (Pr:^Umachine) V J p q =
    LEADSTO Pr V J p q /\  UNLESS Pr V J p q` ;

(*-------------------------------------------------------------------
   some basic laws for STABLE
---------------------------------------------------------------------*)

val STABLE_CONJ = store_thm
  ("STABLE_CONJ",
 --`STABLE (Pr:^Umachine) J1 /\
    STABLE Pr  J2 
    ==>
    STABLE Pr  (J1 AND J2)`--,
    
    PRED_TAC_v3 [STABLE_def, HOA_def] []) ;

val STABLE_EXTEND = store_thm
  ("STABLE_EXTEND",
 --` STABLE (Pr:^Umachine) J1 /\
     (!a::Pr. HOA a (J1 AND J2) J2)
     ==>
     STABLE Pr (J1 AND J2)`--,
     
     PRED_TAC_v3 [STABLE_def, HOA_def] [] ) ;

val STABLE_COMPO = store_thm
  ("STABLE_COMPO",
 --`STABLE (Pr:^Umachine) J /\ STABLE Qr J
    ==>
    STABLE (Pr OR Qr) J`--,
     
     PRED_TAC_v3 [STABLE_def, HOA_def] [] ) ;

(*-------------------------------------------------------------------
   some basic laws for UNLESS
---------------------------------------------------------------------*)

val UNLESS_adef1 = store_thm
  ("UNLESS_adef1",
   --`UNLESS (Pr:^Umachine) V J p q =
      STABLE Pr J /\
      (!a::Pr. HOA a (J AND (p pj V) AND NOT(q pj V)) 
                     (J AND ((p pj V) OR (q pj V))))`--,
   
    PRED_TAC_v3 [UNLESS_def, STABLE_def, HOA_def] []) ;
  

(*-- list of rewrites to push terms of the form (p AND q) pj V to
     (p pj V) AND (q pj V) 
---- *)

val pj_transform = [pj_AND, pj_OR, pj_NOT, pj_IMP, pj_EQUAL, 
                    pj_qOR, pj_qAND, pj_qAND_GEN, pj_qOR_GEN] ;

val UNLESS_SUBST = store_thm
  ("UNLESS_SUBST",
   --`Pvalid V J (p EQUAL r) /\
      Pvalid V J (q EQUAL s) /\
      UNLESS (Pr:^Umachine) V J p q
      ==>
      UNLESS Pr V J r s`--,
   
   (* note : using the alternative definition UNLESS_adef1 is necessary,
             or else PROVE_TAC fails *)

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


   
val UNLESS_CONSQ_WEAK = store_thm
  ("UNLESS_CONSQ_WEAK",
   --`Pvalid V J (q IMP r) /\
      UNLESS (Pr:^Umachine) V J p q
      ==>
      UNLESS Pr V J p r`--,
   
   (* note : using the alternative definition UNLESS_adef1 is necessary,
             or else PROVE_TAC fails *)
   
   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;



val UNLESS_IMP_LIFT1 = store_thm
  ("UNLESS_IMP_LIFT1",
   --`Pvalid V J (p IMP q) /\
      STABLE (Pr:^Umachine) J
      ==>
      UNLESS Pr V J p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val UNLESS_IMP_LIFT2 = store_thm
  ("UNLESS_IMP_LIFT2",
   --`Pvalid V J ((NOT p) IMP q) /\
      STABLE (Pr:^Umachine) J
      ==>
      UNLESS Pr V J p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val UNLESS_CONJ = store_thm
  ("UNLESS_CONJ",
   --`UNLESS (Pr:^Umachine) V J p q /\
      UNLESS Pr V J r s 
      ==>
      UNLESS Pr V J 
             (p AND r) 
             ((p AND s) OR (r AND q) OR (q AND s))`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;
 


val UNLESS_DISJ = store_thm
  ("UNLESS_DISJ",
   --`UNLESS (Pr:^Umachine) V J p q /\
      UNLESS Pr V J r s 
      ==>
      UNLESS Pr V J 
             (p OR r) 
             ((NOT p AND s) OR (NOT r AND q) OR (q AND s))`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;
 

val UNLESS_STABLE_STRONG =  store_thm
  ("UNLESS_STABLE_STRONG",
 --`UNLESS (Pr:^Umachine) V J1 p q /\
    STABLE Pr J2
    ==>
    UNLESS Pr V (J1 AND J2) p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;

val UNLESS_STABLE_STRONG2 =  store_thm
  ("UNLESS_STABLE_STRONG2",
 --`UNLESS (Pr:^Umachine) V J1 p q /\
    STABLE Pr (J1 AND J2)
    ==>
    UNLESS Pr V (J1 AND J2) p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;

val UNLESS_STABLE_LSHIFT =  store_thm
  ("UNLESS_STABLE_LSHIFT",
 --`UNLESS (Pr:^Umachine) V J (p AND q) r /\
    STABLE Pr (J AND (p pj V))
    ==>
    UNLESS Pr V (J AND (p pj V)) q r`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val UNLESS_STABLE_RSHIFT =  store_thm
  ("UNLESS_STABLE_RSHIFT",
 --`UNLESS (Pr:^Umachine) V (J AND (p pj V)) q r /\
    STABLE Pr J
    ==>
    UNLESS Pr V J (p AND q) (p AND r)`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val UNLESS_STABLE_CONJ =  store_thm
  ("UNLESS_STABLE_CONJ",
 --`UNLESS (Pr:^Umachine) V J1 p q /\
    UNLESS  Pr V J2 p q 
    ==>
    UNLESS Pr V (J1 AND J2) p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;
   

val UNLESS_STABLE_DISJ =  store_thm
  ("UNLESS_STABLE_DISJ",
 --`UNLESS (Pr:^Umachine) V J1 p q /\
    UNLESS  Pr V J2 p q 
    ==>
    UNLESS Pr V (J1 OR J2) p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val UNLESS_COMPO =  store_thm
  ("UNLESS_COMPO",
 --`UNLESS (Pr:^Umachine) V J p q /\
    UNLESS  Qr V J p q 
    ==>
    UNLESS (Pr OR Qr) V J p q`--,

   PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def] @ pj_transform) []) ;


val pj_COMPO_lem1 = prove
  ( --`((p:^Pred) pj V) pj (V OR W) = p pj V`--,
    SUBGOAL_THEN (--`(V:^VARS) AND (V OR W) = V`--)
            ASSUME_TAC
    THENL [ PRED_TAC_v3 [] [], ALL_TAC]
    THEN PROVE_TAC[pj_COMPO]) ;

    
val UNLESS_V_EXTEND =  store_thm
  ("UNLESS_V_EXTEND",
 --`UNLESS (Pr:^Umachine) V J p q 
    ==>
    UNLESS Pr (V OR W) J (p pj V) (q pj V)`--,

    PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def, pj_COMPO_lem1]) []) ;
    

val UNLESS_WD_COMPO =  store_thm
  ("UNLESS_WD_COMPO",
 --`UNLESS (Pr:^Umachine) V J p q /\
    STABLE Qr J                   /\
    IGNORE Qr J V  
    ==>
    UNLESS (Pr OR Qr) V J p q`--,

    PRED_TAC_v3 ([UNLESS_adef1, STABLE_def, HOA_def, IGNORE_def]) []) ;

val UNLESS_pj = store_thm
  ("UNLESS_pj",
 --`UNLESS (Pr:^Umachine) V J p q
    =
    UNLESS Pr V J (p pj V) (q pj V)`--,

    REWRITE_TAC [UNLESS_def, pj_pj]) ;

val UNLESS_SINGH = store_thm
  ("UNLESS_SINGH",
 --`UNLESS (Pr:^Umachine) V J p q /\
    STABLE Qr J /\
    CONDWRITE  Qr J V (r pj (V OR W)) (s pj (V OR W))  
    ==>
    UNLESS (Pr OR Qr) (V OR W) J
           ((p pj V) AND r)
           (((q pj V) AND r) OR (NOT r) OR s)`--,

    PRED_TAC_v3 ([UNLESS_def, STABLE_def, HOA_def, CONDWRITE_def,
                  pj_COMPO_lem1] @ pj_transform) []) ;


(*-------------------------------------------------------------------
   some basic laws for ENSURES
---------------------------------------------------------------------*)

val ENSURES_adef1 = store_thm
  ("ENSURES_adef1",
   --`ENSURES (Pr:^Umachine) V J p q =
      UNLESS Pr V J p q /\ 
      (?a::Pr. HOA a (J AND (p pj V) AND NOT(q pj V)) 
                     (J AND (q pj V)))`--,
   
    PRED_TAC_v3 [ENSURES_def, UNLESS_def, STABLE_def, HOA_def] []) ;


val ENSURES_IMP_LIFT = store_thm
  ("ENSURES_IMP_LIFT",
   --`Pr a /\
      Pvalid V J (p IMP q) /\
      STABLE (Pr:^Umachine) J 
      ==>
      ENSURES Pr V J p q`--,

    PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                  @ pj_transform) []) ;


val ENSURES_CONSQ_WEAK = store_thm
  ("ENSURES_CONSQ_WEAK",
   --`Pvalid V J (q IMP r) /\
      ENSURES (Pr:^Umachine) V J p q
      ==>
      ENSURES Pr V J p r`--,

   PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                  @ pj_transform) []) ;

(*--
   original proof :

   REWRITE_TAC [ENSURES_adef1]
   THEN REPEAT STRIP_TAC 
   THENL [ PROVE_TAC [UNLESS_CONSQ_WEAK] , ALL_TAC ]
   THEN UNDISCH_ALL_TAC
   THEN REWRITE_TAC [Pvalid_def, SYM pj_IMP, pSEQ_DEF]
   THEN RESTRICT_DEF_TAC
   THEN REWRITE_TAC [HOA_def, AND_def, NOT_def, OR_def, IMP_def] 
   THEN BETA_TAC 
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC (--`x:^Action`--)
   THEN PROVE_TAC[]) ;

     note : it is peculier that in above theorem the meson prover
     needs a bit help (the EXISTS_TAC (--`x:^Action`--)).
     I have asked to John Harrison and waiting for answer.
     The problem is not clear ... maybe it has problem with
     the second order quantifier. Or maybe it's simply a bug.

     A simplified goal looks like this :

     val yoo = 
     --`Pr x /\ 
     (!(s:'a) (t:'a). J s /\ P s  /\ x s t ==> J t /\ Q t) /\ 
     (!z. J z /\ Q z ==> R z)
     ==> 
     (?x. Pr x /\ (!s t. J s /\ P s  /\ x s t ==> J t /\ R t))`-- ;

     Try it ...
  
     prove(yoo, PROVE_TAC[]) --> fail, but :
     prove(yoo, STRIP_TAC THEN EXISTS_TAC (--`x:'a->'a->bool`--)
                THEN PROVE_TAC[]) --> ok

     but also :
     val yuu = 
     --`Pr x /\ 
     (!(s:'a) (t:'a). P s  /\ x s t ==> Q t) /\ 
     (!z. Q z ==> R z)
     ==> 
     (?x. Pr x /\ (!s t. P s  /\ x s t ==> R t))`-- ;

--*)

val ENSURES_PSP = store_thm
  ("ENSURES_PSP",
   --`ENSURES (Pr:^Umachine) V J p q /\
      UNLESS Pr V J r s
      ==>
      ENSURES Pr V J (p AND r) ((q AND r) OR s)`--,

    PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                  @ pj_transform) []
    THEN mesonLib.GEN_MESON_TAC 0 40 1 []) ;


val ENSURES_STABLE_STRONG =  store_thm
  ("ENSURES_STABLE_STRONG",
 --`ENSURES (Pr:^Umachine) V J1 p q /\
    STABLE Pr J2
    ==>
    ENSURES Pr V (J1 AND J2) p q`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                [] ) ;


val ENSURES_STABLE_STRONG2 =  store_thm
  ("ENSURES_STABLE_STRONG2",
 --`ENSURES (Pr:^Umachine) V J1 p q /\
    STABLE Pr (J1 AND J2)
    ==>
    ENSURES Pr V (J1 AND J2) p q`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                [] ) ;

val ENSURES_STABLE_LSHIFT =  store_thm
  ("ENSURES_STABLE_LSHIFT",
 --`ENSURES (Pr:^Umachine) V J (p AND q) r /\
    STABLE Pr (J AND (p pj V))
    ==>
    ENSURES Pr V (J AND (p pj V)) q r`--,

    PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                  @ pj_transform)
                [] 
    THEN mesonLib.GEN_MESON_TAC 0 40 1 []) ;


val ENSURES_STABLE_RSHIFT =  store_thm
  ("ENSURES_STABLE_RSHIFT",
 --`ENSURES (Pr:^Umachine) V (J AND (p pj V)) q r /\
    STABLE Pr J
    ==>
    ENSURES Pr V J (p AND q) (p AND r)`--,

    
    PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                  @ pj_transform)
                [] 
    THEN mesonLib.GEN_MESON_TAC 0 40 1 []) ;


val ENSURES_STABLE_CONJ =  store_thm
  ("ENSURES_STABLE_CONJ",
 --`ENSURES (Pr:^Umachine) V J1 p q /\
    ENSURES  Pr V J2 p q 
    ==>
    ENSURES Pr V (J1 AND J2) p q`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                [] ) ;


(*-- note : ENSURES_STABLE_DISJ cannot be proven. This is understandable
     since if we know J1 |- p ens q and J2 |- p ens q we only know that'
     there is an action a guaranteeing progress form J1 to q, and another
     action b guaranteeing progress from J2 to q, but if J1 OR J2 holds, 
     then we don't know if there is a single action that guarantees q 
     (unless a and b are the same, which we also don't know) --*)


val ENSURES_COMPO =  store_thm
  ("ENSURES_COMPO",
 --`ENSURES (Pr:^Umachine) V J p q /\
    UNLESS  Qr V J p q 
    ==>
    ENSURES (Pr OR Qr) V J p q`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def] 
                [] ) ;


val ENSURES_V_EXTEND =  store_thm
  ("ENSURES_V_EXTEND",
 --`ENSURES (Pr:^Umachine) V J p q 
    ==>
    ENSURES Pr (V OR W) J (p pj V) (q pj V)`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def,
                 pj_COMPO_lem1] 
                [] ) ;


val ENSURES_WD_COMPO =  store_thm
  ("ENSURES_WD_COMPO",
 --`ENSURES (Pr:^Umachine) V J p q /\
    STABLE Qr  J /\
    IGNORE  Qr J V  
    ==>
    ENSURES (Pr OR Qr) V J p q`--,

    PRED_TAC_v3 [ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def,
                 IGNORE_def] 
                [] ) ;

val ENSURES_pj = store_thm
  ("ENSURES_pj",
 --`ENSURES (Pr:^Umachine) V J p q
    =
    ENSURES Pr V J (p pj V) (q pj V)`--,

    REWRITE_TAC [ENSURES_def, UNLESS_def, pj_pj]) ;  

val ENSURES_SINGH = store_thm
  ("ENSURES_SINGH",
 --`ENSURES (Pr:^Umachine) V J p q /\
    STABLE Qr J /\
    CONDWRITE  Qr J V (r pj (V OR W)) (s pj (V OR W))  
    ==>
    ENSURES (Pr OR Qr) (V OR W) J
            ((p pj V) AND r)
            (((q pj V) AND r) OR (NOT r) OR s)`--,

    PRED_TAC_v3 ([ENSURES_adef1, UNLESS_adef1, STABLE_def, HOA_def,
                  CONDWRITE_def, pj_COMPO_lem1] @ pj_transform)
                [] 
    THEN mesonLib.GEN_MESON_TAC 0 40 1 []) ;

(*-------------------------------------------------------------------
  Here are basic LEADSTO laws as in CM88, plus :

    * the pre-strengthening and post-weakening forms an equivalent
      of Sander's style of substitution law for leadsto.
    * progress now implies J stability and non-emptiness of the
      program 
---------------------------------------------------------------------*)

val LEADSTO_INDUCT_thm1 = store_thm
  ("LEADSTO_INDUCT_thm1",
   --`((ENSURES (Pr:^Umachine) V J) SUBREL U) /\
      TRANS U /\
      LDISJ U
      ==>
      (!p q. LEADSTO Pr V J p q ==> U p q)`--,

    REWRITE_TAC [LEADSTO_def, SUBREL_def, 
                 REWRITE_RULE [SUBREL_def] TDC_INDUCT1]) ;
    
val LEADSTO_ENS_LIFT = store_thm
  ("LEADSTO_ENS_LIFT",
   --`ENSURES (Pr:^Umachine) V J p q 
      ==>
      LEADSTO Pr V J p q`--,
   
    REWRITE_TAC [LEADSTO_def, REWRITE_RULE [SUBREL_def] TDC_LIFT] ) ;


val LEADSTO_TRANS = store_thm
  ("LEADSTO_TRANS",
   --`LEADSTO (Pr:^Umachine) V J p q /\
      LEADSTO Pr V J q r
      ==>
      LEADSTO Pr V J p r`--,
   
    REWRITE_TAC [LEADSTO_def, REWRITE_RULE [TRANS_def] TDC_TRANS] ) ;


val LEADSTO_LDISJ = store_thm
  ("LEADSTO_LDISJ",
 --`W p /\
    (!p::W. LEADSTO (Pr:^Umachine) V J p q)
    ==>
    LEADSTO Pr V J (??p::W. p) q`--,
   
    REWRITE_TAC [LEADSTO_def]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [LDISJ_def] TDC_LDISJ)
    THEN ASM_REWRITE_TAC [] THEN PROVE_TAC[]) ;
   

val LEADSTO_IMP_LIFT = store_thm
  ("LEADSTO_IMP_LIFT",
   --`Pr a /\
      Pvalid V J (p IMP q ) /\
      STABLE (Pr:^Umachine) J 
      ==>
      LEADSTO Pr V J p q`--,

      PROVE_TAC [ENSURES_IMP_LIFT, LEADSTO_ENS_LIFT]) ;
    
fun MK_ABS2_CONV x y trm = 
    (SYM o DEPTH_CONV BETA_CONV o Term) `(\^x ^y. ^trm) ^x ^y` ; 

(*-----------------------------------------------------------------------
 
  Here is is an induction tactic. 

  LEADSTO_INDUCT_v1 lems_e lems_t lems_d do the following :

    (1) find the induction variables, and automate the lamda lifting
        and subsequent induction application
    (2) meson prover is invoked at each case (ensures lift, transitivity,
        and disjunction) to solve them. The three additional arguments 
        are lists of theorems which are passed to the meson prover to
        prove its goal.
-----------------------------------------------------------------------*)

fun LEADSTO_INDUCT_TAC_v1 lems_e lems_t lems_d (asml,g) =
    let    
    val leadsto_term =
        let 
        val ante = (rand o rator) g
        in
        if is_conj ante then (rand o rator) ante
                              else ante
        end
    val p_term = (rand o rator) leadsto_term 
    val q_term = rand leadsto_term
    in
    (REPEAT GEN_TAC
     THEN STRIP_TAC
     THEN CONV_TAC (MK_ABS2_CONV p_term q_term)
     THEN UNDISCH_TAC leadsto_term
     THEN MATCH_MP_TAC LEADSTO_INDUCT_thm1     
     THEN CONJ_TAC 
     THENL
     [ (*-- ensures part --*)
       REWRITE_TAC [SUBREL_def, restr_binderTheory.RES_FORALL,
                    restr_binderTheory.RES_EXISTS ]
       THEN BETA_TAC
       THEN REPEAT STRIP_TAC 
       THEN TRY (PROVE_TAC lems_e),
       CONJ_TAC
       THENL
       [ (*-- transitivity part --*)
         REWRITE_TAC [TRANS_def, restr_binderTheory.RES_FORALL,
                       restr_binderTheory.RES_EXISTS ]
         THEN BETA_TAC
         THEN REPEAT STRIP_TAC 
         THEN TRY (PROVE_TAC lems_t),
         (*-- disjunctivity part --*)
         REWRITE_TAC [LDISJ_def, restr_binderTheory.RES_FORALL,
                      restr_binderTheory.RES_EXISTS ]
         THEN BETA_TAC 
         THEN REPEAT STRIP_TAC 
         THEN TRY (PROVE_TAC lems_d)
       ]
     ])
     (asml,g)
     end ;
        

val LEADSTO_IMP_STABLE = store_thm
  ("LEADSTO_IMP_STABLE",
   --`LEADSTO (Pr:^Umachine) V J p q 
      ==>
      STABLE Pr J`--,
   
   LEADSTO_INDUCT_TAC_v1 [ENSURES_def, UNLESS_def] [] [] ) ;


val LEADSTO_IMP_NONEMPTY = store_thm
  ("LEADSTO_IMP_NONEMPTY",
   --`LEADSTO (Pr:^Umachine) V J p q 
      ==>
      (?a. Pr a)`--,

   LEADSTO_INDUCT_TAC_v1 [RESTRICT_DEF_RULE ENSURES_def] [] [] ) ;


val LEADSTO_CONSQ_WEAK = store_thm
  ("LEADSTO_CONSQ_WEAK",
 --`LEADSTO (Pr:^Umachine) V J p q /\
    Pvalid V J (q IMP r) 
    ==>
    LEADSTO Pr V J p r`--,
    
    PROVE_TAC [LEADSTO_IMP_LIFT, 
               LEADSTO_IMP_STABLE, 
               LEADSTO_IMP_NONEMPTY, 
               LEADSTO_TRANS]) ;

val LEADSTO_ANTE_STRONG = store_thm
  ("LEADSTO_ANTE_STRONG",
 --`LEADSTO (Pr:^Umachine) V J q r /\
    Pvalid V J (p IMP q) 
    ==>
    LEADSTO Pr V J p r`--,
    
    PROVE_TAC [LEADSTO_IMP_LIFT, 
               LEADSTO_IMP_STABLE, 
               LEADSTO_IMP_NONEMPTY, 
               LEADSTO_TRANS]) ;

val lemma1 = BETA_RULE
            (SPEC (--`(\z:^Pred. (z=p) \/ (z=r))`--)
            (GEN (--`W:^Pred->bool`--) LEADSTO_LDISJ)) ;

val lemma2 = prove
 (--`(p:^Pred) OR r = (??p::(\z. (z=p)\/(z=r)). p)`--,
  PRED_TAC_v3 [] []) ;


val LEADSTO_LDISJ2 = store_thm
  ("LEADSTO_LDISJ2",
 --`LEADSTO (Pr:^Umachine) V J p s /\
    LEADSTO Pr V J r s 
    ==>
    LEADSTO Pr V J (p OR r) s`--,
    
    REWRITE_TAC [lemma2]
    THEN STRIP_TAC    
    THEN MATCH_MP_TAC lemma1 
    THEN RESTRICT_DEF_TAC
    THEN PROVE_TAC[]) ;

val LEADSTO_DISJ2 = store_thm
  ("LEADSTO_DISJ2",
 --`LEADSTO (Pr:^Umachine) V J p q /\
    LEADSTO Pr V J r s 
    ==>
    LEADSTO Pr V J (p OR r) (q OR s)`--,
    
    STRIP_TAC
    THEN PROVE_TAC 
         [LEADSTO_CONSQ_WEAK, 
          LEADSTO_LDISJ2,
          prove (--`Pvalid V (J:^Pred) (q IMP (q OR s))`--,
                 PRED_TAC_v3 pj_transform []),
          prove (--`Pvalid V (J:^Pred) (s IMP (q OR s))`--,
                 PRED_TAC_v3 pj_transform [])]
    ) ;
 

val LEADSTO_REFL =  store_thm
  ("LEADSTO_REFL",
 --`Pr a  /\
    STABLE (Pr:^Umachine) J
    ==>
    LEADSTO Pr V J p p`--,
   
    STRIP_TAC
    THEN PROVE_TAC 
         [LEADSTO_IMP_LIFT,
          prove (--`Pvalid V (J:^Pred) (p IMP p)`--,
                 PRED_TAC_v3 pj_transform []) ]
     ) ;


val LEADSTO_ty = ty_antiq (Type `:^Umachine->^VARS->
                                  ^Pred->^Pred->^Pred->bool`) ;

val lemma = prove
 (--`(??i::TT. (f:bool->^Pred) i) = (f T) OR (f F)`--,
     PRED_TAC_v3 [] [] 
     THEN EQ_TAC THEN REPEAT STRIP_TAC
     THENL
     [ ASM_CASES_TAC (--`T=i`--)
             THENL [ FIRST_ASSUM SUBST1_TAC THEN PROVE_TAC[],
                     SUBGOAL_THEN (--`F=i`--) SUBST1_TAC
                     THEN PROVE_TAC[] ],
       EXISTS_TAC (--`T`--) THEN ASM_REWRITE_TAC[],
       EXISTS_TAC (--`F`--) THEN ASM_REWRITE_TAC[] 
     ] ) ;


val LEADSTO_TT_DISJ = prove
 (--`DISJ (TT:bool->bool) 
          ((LEADSTO:^LEADSTO_ty) Pr V J)`--,
    REWRITE_TAC [DISJ_def]
    THEN RESTRICT_DEF_TAC
    THEN REWRITE_TAC [lemma]
    THEN REWRITE_TAC [TT_def]
    THEN PROVE_TAC [LEADSTO_DISJ2]) ;


val LEADSTO_LDISJ_GEN = store_thm
  ("LEADSTO_LDISJ_GEN",
 --`W (i:'index) /\
    (!i::W. (LEADSTO:^LEADSTO_ty) Pr V J (f i) q)
    ==>
    LEADSTO Pr V J (??i::W. f i) q`--,
    
    REWRITE_TAC [LEADSTO_def]
    THEN STRIP_TAC THEN MATCH_MP_TAC LDISJ_GEN
    THEN PROVE_TAC [TDC_LDISJ]) ;

val LEADSTO_DISJ_GEN = store_thm
  ("LEADSTO_DISJ_GEN",
 --`W (i:'index) /\
    (!i::W. LEADSTO (Pr:^Umachine) V J (f i) (g i))
    ==>
    LEADSTO Pr V J (??i::W. f i) (??i::W. g i)`--,

    RESTRICT_DEF_TAC
    THEN STRIP_TAC
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE LEADSTO_LDISJ_GEN)
    THEN REPEAT STRIP_TAC
    THEN TRY (PROVE_TAC[])
    THEN SUBGOAL_THEN (--`Pvalid V J (((g:'index->^Pred) x)
                                       IMP
                                       (??i::W. g i))`--)
         ASSUME_TAC
    THENL [PRED_TAC_v3 pj_transform [], 
           PROVE_TAC [LEADSTO_CONSQ_WEAK] ]
   ) ;
                   

val MRel   = ty_antiq (Type `:'m->'m->bool`) ;
val Metric = ty_antiq (Type `:^State->'m`) ;

val LEADSTO_BOUNDED_PROG = store_thm
  ("LEADSTO_BOUNDED_PROG",
 --`ADMIT_WF_INDUCTION LESS /\
    (!m. (LEADSTO:^LEADSTO_ty) 
          Pr V J
          (p AND (\s. (M:^Metric) s = m)) 
          ((p AND (\s. LESS (M s) m)) OR q))
     ==>
     LEADSTO Pr V J p q`--,
    
     STRIP_TAC
     THEN MATCH_MP_TAC BOUNDED_REACH
     THEN (EXISTSL_TAC [--`LESS:^MRel`--,
                        --`M:^Metric`--   ])
     THEN ASM_REWRITE_TAC [TRANS_def, LDISJ_def,
                           LEADSTO_TRANS,
                           LEADSTO_TT_DISJ ]
     THEN REPEAT STRIP_TAC
     THENL
       [ MATCH_MP_TAC (GEN_ALL LEADSTO_LDISJ)
         THEN ASM_REWRITE_TAC[] 
         THEN PROVE_TAC [],
         PROVE_TAC [LEADSTO_LDISJ,
                     LEADSTO_IMP_STABLE,
                     LEADSTO_IMP_NONEMPTY,
                     LEADSTO_REFL]
       ]
   ) ;

val  LEADSTO_CANCEL = store_thm
  ("LEADSTO_CANCEL",
 --`(LEADSTO:^LEADSTO_ty) Pr V J p (q OR r) /\
     LEADSTO              Pr V J r  s       
     ==>
     LEADSTO Pr V J p (q OR s)`--,
    
     STRIP_TAC
     THEN MATCH_MP_TAC CANCEL
     THEN ASM_REWRITE_TAC [TRANS_def, LDISJ_def,
                           LEADSTO_TRANS, LEADSTO_LDISJ,
                           LEADSTO_TT_DISJ ]
     THEN PROVE_TAC [LEADSTO_IMP_STABLE,
                     LEADSTO_IMP_NONEMPTY,
                     LEADSTO_REFL]
   ) ;
     

val LEADSTO_PSP = store_thm
  ("LEADSTO_PSP",
 --`(LEADSTO:^LEADSTO_ty) Pr V J p q /\
    UNLESS Pr V J r s
    ==>
    LEADSTO Pr V J (p AND r) ((q AND r) OR s)`--,


    LEADSTO_INDUCT_TAC_v1 
         [ENSURES_PSP, LEADSTO_ENS_LIFT]
         [LEADSTO_CANCEL,
          ISPECL [--`(y:^Pred) AND r`--, --`s:^Pred`--] pOR_SYM,
          prove(--`(z:^Pred) AND r OR s = s OR (z AND r OR s)`--,
                PRED_TAC_v3 [] [])]
         []
    (*-- disjunctivity part --*)
    THEN SUBST1_TAC
         (prove(--`(??(x:^Pred)::W. x) AND r = 
                   (??x::W. (\i. i AND r) x)`--,
                PRED_TAC_v3 [] []))
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
    THEN ASM_REWRITE_TAC[]
    THEN  PROVE_TAC[] ) ;



val LEADSTO_STABLE_STRONG = store_thm
  ("LEADSTO_STABLE_STRONG",
 --`(LEADSTO:^LEADSTO_ty) Pr V J1 p q /\
    STABLE Pr J2
    ==>
    LEADSTO Pr V (J1 AND J2) p q`--,

    LEADSTO_INDUCT_TAC_v1 
      [ENSURES_STABLE_STRONG, LEADSTO_ENS_LIFT]
      [LEADSTO_TRANS]
      []
    (*-- disjunction case --*)
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ))
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC[]) ;

val LEADSTO_STABLE_STRONG2 = store_thm
  ("LEADSTO_STABLE_STRONG2",
 --`LEADSTO (Pr:^Umachine) V J1 p q /\
    STABLE Pr (J1 AND J2)
    ==>
    LEADSTO Pr V (J1 AND J2) p q`--,

    LEADSTO_INDUCT_TAC_v1 
      [ENSURES_STABLE_STRONG2, LEADSTO_ENS_LIFT]
      [LEADSTO_TRANS]
      []
    (*-- disjunction case --*)
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ))
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC[]) ;
 

val LEADSTO_STABLE_RSHIFT = store_thm
  ("LEADSTO_STABLE_RSHIFT",
 --`(LEADSTO:^LEADSTO_ty) Pr V (J AND (p pj V)) q r /\
    STABLE Pr J
    ==>
    LEADSTO Pr V J (p AND q) (p AND r)`--,

    LEADSTO_INDUCT_TAC_v1 
      [ENSURES_STABLE_RSHIFT, LEADSTO_ENS_LIFT]
      [LEADSTO_TRANS]
      []
    (*-- disjunction case --*)
    THEN SUBST1_TAC
           (prove(--`p AND (??(x:^Pred)::W. x)  = 
                     (??x::W. (\i. p AND i) x)`--,
                  PRED_TAC_v3 [] []))
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC[] ) ;

 
val LEADSTO_STABLE_LSHIFT = store_thm
  ("LEADSTO_STABLE_LSHIFT",
 --`(LEADSTO:^LEADSTO_ty) Pr V J (p AND q) r /\
    STABLE Pr (J AND (p pj V))
    ==>
    LEADSTO Pr V (J AND (p pj V)) q r`--,
  
    PROVE_TAC
      [ prove (--`(J:^Pred) AND (J AND (p pj V)) = J AND (p pj V)`--,
               PRED_TAC_v3 [] []),
        prove (--`Pvalid (V:^VARS) ((J:^Pred) AND (p pj V))
                                   (q IMP  p AND q)`--,
               PRED_TAC_v3 ([pj_pj] @ pj_transform) []),

        LEADSTO_STABLE_STRONG, 
        LEADSTO_ANTE_STRONG ]
   ) ;


val LEADSTO_V_EXTEND =  store_thm
  ("LEADSTO_V_EXTEND",
 --`LEADSTO (Pr:^Umachine) V J p q 
    ==>
    LEADSTO Pr (V OR W) J (p pj V) (q pj V)`--,


    LEADSTO_INDUCT_TAC_v1 
      [ENSURES_V_EXTEND, LEADSTO_ENS_LIFT]
      [LEADSTO_TRANS]
      []
    (*-- disjucntion case --*)
    THEN REWRITE_TAC [pj_qOR]
    THEN SUBST1_TAC
        (prove(--`(??(i:^Pred)::W'. i pj V)  = 
                  (??i::W'. (\x. x pj V) i)`--,
                  PRED_TAC_v3 [] []))
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC[] ) ;


val LEADSTO_WD_COMPO =  store_thm
  ("LEADSTO_WD_COMPO",
 --`LEADSTO (Pr:^Umachine) V J p q /\
    STABLE Qr J /\
    IGNORE Qr J V  
    ==>
    LEADSTO (Pr OR Qr) V J p q`--,

    LEADSTO_INDUCT_TAC_v1 
      [ENSURES_WD_COMPO, LEADSTO_ENS_LIFT]
      [LEADSTO_TRANS]
      []
    (*-- disjucntion case --*)
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ))
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC []) ;


val LEADSTO_SINGH =  store_thm
  ("LEADSTO_SINGH",
 --`LEADSTO (Pr:^Umachine) V J p q /\
    STABLE Qr J /\
    CONDWRITE  Qr J V (r pj (V OR W)) (s pj (V OR W))  
    ==>
    LEADSTO (Pr OR Qr) (V OR W) J
            ((p pj V) AND r)
            (((q pj V) AND r) OR (NOT r) OR s)`--,

  LEADSTO_INDUCT_TAC_v1 
      [ENSURES_SINGH, LEADSTO_ENS_LIFT]
      []
      []
  THENL
  [ (*-- transitivity case --*)
    UNDISCH_ALL_TAC
    THEN ONCE_REWRITE_TAC [pOR_SYM]
    THEN REPEAT STRIP_TAC
    THEN ONCE_REWRITE_TAC 
          [ prove(--`(NOT r OR s) OR ((z pj V) AND r)
                      =
                     (NOT (r:^Pred) OR s) OR ((z pj V) AND r) OR
                     (NOT r OR s)`--,
                  PRED_TAC_v3 [] [])]
    THEN MATCH_MP_TAC (GEN_ALL LEADSTO_CANCEL)
    THEN EXISTS_TAC (--`(y pj V) AND (r:^Pred)`--) 
    THEN PROVE_TAC [pOR_SYM],

    (*-- disjunctivity case --*)
     REWRITE_TAC [pj_qOR]
     THEN SUBST1_TAC
           (prove(--`(??(i:^Pred)::W'. i pj V) AND r = 
                     (??i::W'. (\x. (x pj V) AND r) i)`--,
                  PRED_TAC_v3 [] []))
     THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
     THEN ASM_REWRITE_TAC[]
     THEN PROVE_TAC[]
    ]
  ) ;


val lemma0 = prove
 (--`LEADSTO (Pr:^Umachine) V J p q
    ==>
    LEADSTO Pr V J (p pj V) (q pj V)`--,

    LEADSTO_INDUCT_TAC_v1 
         [LEADSTO_ENS_LIFT, ENSURES_pj] 
         [LEADSTO_TRANS] 
         []
    THEN REWRITE_TAC [pj_qOR]
    THEN SUBST1_TAC
           (prove(--`(??(i:^Pred)::W. i pj V) = 
                     (??i::W. (\x. x pj V) i)`--,
                  PRED_TAC_v3 [] []))
     THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
     THEN ASM_REWRITE_TAC[]
     THEN PROVE_TAC[]) ;

val lemma1 = prove
(--`Pr a /\ STABLE (Pr:^Umachine) J ==> ENSURES Pr V J p (p pj V)`--,

    STRIP_TAC
    THEN MATCH_MP_TAC ENSURES_IMP_LIFT
    THEN ASM_REWRITE_TAC [Pvalid_def,pj_IMP,pj_pj]
    THEN PRED_TAC_v3 [] []) ;

val lemma2 = prove
(--`Pr a /\ STABLE (Pr:^Umachine) J ==> ENSURES Pr V J (p pj V) p`--,

    STRIP_TAC
    THEN MATCH_MP_TAC ENSURES_IMP_LIFT
    THEN ASM_REWRITE_TAC [Pvalid_def,pj_IMP,pj_pj]
    THEN PRED_TAC_v3 [] []) ;


val lemma3 = prove
(--`LEADSTO (Pr:^Umachine) V J (p pj V) (q pj V)
    ==>
    LEADSTO Pr V J p q`--,

    STRIP_TAC
    THEN SUBGOAL_THEN (--`LEADSTO (Pr:^Umachine) V J p (p pj V)`--)
         ASSUME_TAC
    THENL [ PROVE_TAC [lemma1, LEADSTO_ENS_LIFT,
                       LEADSTO_IMP_STABLE,
                       LEADSTO_IMP_NONEMPTY ],
            ALL_TAC ]
    THEN SUBGOAL_THEN (--`LEADSTO (Pr:^Umachine) V J (q pj V) q`--)
         ASSUME_TAC
    THENL [ PROVE_TAC [lemma2, LEADSTO_ENS_LIFT,
                       LEADSTO_IMP_STABLE,
                       LEADSTO_IMP_NONEMPTY ],
            ALL_TAC ]
    THEN PROVE_TAC [LEADSTO_TRANS]) ;

val LEADSTO_pj = store_thm
  ("LEADSTO_pj",
 --`LEADSTO (Pr:^Umachine) V J p q
    =
    LEADSTO Pr V J (p pj V) (q pj V)`--,

   PROVE_TAC [lemma0, lemma3]) ;


val lemma0 = 
 (REWRITE_RULE [prove (--`(q:^Pred) AND p OR NOT p OR q = q OR NOT p`--,
                PRED_TAC_v3 [] [] ) ] o
  REWRITE_RULE [SYM LEADSTO_pj] o
  REWRITE_RULE (map SYM pj_transform) o
  REWRITE_RULE [pAND_REFL, pj_pj] o
  REWRITE_RULE pj_transform o
  ONCE_REWRITE_RULE [LEADSTO_pj] o
  REWRITE_RULE [pOR_REFL] o
  ISPECL [--`q:^Pred`--,
          --`V:^VARS`--,
          --`p:^Pred`--,
          --`Qr:^Umachine`--,
          --`q:^Pred`--,
          --`p:^Pred`--,
          --`J:^Pred`--,
          --`V:^VARS`--,
          --`Pr:^Umachine`--] o
  GEN_ALL) 
  LEADSTO_SINGH ;

val lemma1 = prove
(--`STABLE (Pr:^Umachine) J /\
    CONDWRITE Pr J V (p pj V) (q pj V) 
    ==>
    UNLESS Pr V J p q`--,

    PRED_TAC_v3 [CONDWRITE_def, UNLESS_def, HOA_def]
                ([pj_pj] @ pj_transform)
                 ) ;

val UNTIL_COMPO = store_thm
  ("UNTIL_COMPO",
 --`UNTIL (Pr:^Umachine) V J p q /\
    STABLE Qr J /\
    CONDWRITE  Qr J V (p pj V) (q pj V)
    ==>
    UNTIL (Pr OR Qr) V J p q`--,

    REWRITE_TAC [UNTIL_def] THEN REPEAT STRIP_TAC
    THENL
    [ (*-- the leadsto part --*)
      by (`LEADSTO ((Pr:^Umachine) OR Qr) V J p (q OR NOT p)`,
          PROVE_TAC [lemma0])
      THEN by (`UNLESS ((Pr:^Umachine) OR Qr) V J p q`,
               PROVE_TAC [UNLESS_COMPO, lemma1])
      THEN PROVE_TAC [LEADSTO_PSP,
                      ISPECL [--`p:^Pred`--] pAND_REFL, 
                      prove (--`(q OR NOT p) AND p OR q = (q:^Pred)`--,
                             PRED_TAC_v3 [] []) ],
      (*-- the unless part --*)
      PROVE_TAC [UNLESS_COMPO, lemma1]
     ]) ;


(*------------------------------------------------------------
   characterize:
 
     - the guard of an action 
     - enabledness of an action
------------------------------------------------------------*)

val disabledBy_def = Define
   `disabledBy (a:^Action) d 
    =
    !p. HOA a (d AND p) p` ;

val disabledBy_adef1 = store_thm
  ("disabledBy_adef1",
 --`disabledBy (a:^Action) d
    =
    (!s t. d s /\ a s t ==> (t = s))`--,
    
    REWRITE_TAC [disabledBy_def, HOA_def, AND_def]
    THEN BETA_TAC
    THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL
    [ FIRST_ASSUM 
      (fn thm=> PROVE_TAC [BETA_RULE 
                           (SPEC (--`(\z. z = (s:^State))`--) 
                           thm)]),
      PROVE_TAC[] ] ) ;      


val SP_def = Define
   `SP V (a:^Action) (p:^Pred) 
    = 
    (\t. (?s t'. ((t Pj V) = t' Pj V) /\ a s t' /\ p s))` ;


val SP_thm1 = store_thm
  ("SP_thm1",
 --`HOA (a:^Action) p (q pj V)
    ==>
    |== (SP V a p IMP (q pj V))`--,

    PRED_TAC_v3 [HOA_def, SP_def, pj_def] [] ) ;
 
   
val SP_thm2 = store_thm
  ("SP_thm2",
 --`HOA (a:^Action) p (SP V a p)`--,
    
    PRED_TAC_v3 [HOA_def, SP_def, pj_def] []) ;


val SP_thm3 = store_thm
  ("SP_thm3",
 --`(SP V (a:^Action) p) pj V = SP V a p`--,

    PRED_TAC_v3 [SP_def, pj_def] [Pj_Pj]) ;
    

val runSet_def = Define
   `runSet (Pr:^Umachine) guard V J (p:^Pred)
    =
    (\a. Pr a /\ ~|==(J AND (p pj V) IMP NOT(guard a)))` ;


val NEXT_def = Define
   `NEXT (Pr:^Umachine) guard V J p
    =
    RES_qOR (\a. runSet Pr guard V J p a) 
            (\a. SP V a (J AND (p pj V)))` ;


val ProgressSpace_def = Define
   `ProgressSpace (Pr:^Umachine) 
                  (guard:^Action->^Pred) 
                   V 
                   J p
    = 
    RES_qOR (\a. runSet Pr guard V J p a) 
            (\a. J AND (p pj V) AND guard a)` ;


val NEXT_LEADSTO_thm = store_thm
  ("NEXT_LEADSTO_thm",
 --`(Pr:^Umachine) a /\
    STABLE Pr J /\
    (!a. Pr a ==> disabledBy a (NOT (guard a))) /\
    Pvalid V J (p
                IMP
                ProgressSpace Pr guard V J p)
    ==>
    LEADSTO Pr  V J
        p
       (NEXT Pr guard V J p)`--,

    STRIP_TAC
    THEN MATCH_MP_TAC (GEN_ALL LEADSTO_ANTE_STRONG)
    THEN POP_ASSUM 
         (fn thm=> EXISTS_TAC ((rand o rand o concl) thm)
                   THEN REWRITE_TAC [thm])
    THEN REWRITE_TAC [ProgressSpace_def]
    THEN ASM_CASES_TAC (--`runSet (Pr:^Umachine) guard V J (p:^Pred) 
                           = 
                           FF`--)
    THENL
    [ (*-- suppose runableSet is empty --*)
      FIRST_ASSUM (fn thm=> SUBST1_TAC thm) 
      THEN MATCH_MP_TAC LEADSTO_IMP_LIFT
      THEN REPEAT CONJ_TAC
      THEN TRY (PROVE_TAC [])
      THEN PRED_TAC_v3 [pj_def] [] ,
      ALL_TAC ]
    (*-- runableSet is non-empty --*)
    THEN SUBGOAL_THEN
         (--`(\a:^Action. (J:^Pred) AND (p pj V) AND guard a) =
             (\a. (\b. J AND (p pj V) AND guard b) a)`--)
         SUBST1_TAC
    THENL [PRED_TAC_v3 [] [], ALL_TAC]
    THEN MATCH_MP_TAC (RESTRICT_DEF_RULE (GEN_ALL LEADSTO_LDISJ_GEN))
    THEN UNDISCH_ALL_TAC 
    THEN REWRITE_TAC [prove(--`~((Z:^Action->bool)=FF) =
                               (?a. Z a)`--,
                      PRED_TAC_v3 [] [])]
    THEN REPEAT STRIP_TAC
    THEN EXISTS_TAC (--`a':^Action`--)
    THEN BETA_TAC THEN ASM_REWRITE_TAC[]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC LEADSTO_ENS_LIFT
    THEN ASM_REWRITE_TAC [NEXT_def, ENSURES_def, UNLESS_def]     
    THEN SUBGOAL_THEN
         (--`(\a:^Action. SP V a (J AND (p pj V)))
             =
             (\a. (\b. SP V b (J AND (p pj V))) a)`--)
         SUBST1_TAC
    THENL [PRED_TAC_v3 [] [], ALL_TAC]
    THEN REWRITE_TAC pj_transform
    THEN BETA_TAC THEN REWRITE_TAC [SP_thm3]
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THENL
      [ (*-- other actions --*)
        ASM_CASES_TAC (--`runSet (Pr:^Umachine) guard V J (p:^Pred) x'`--)
        THENL
        [ (*-- if the action is runable --*)
           SUBGOAL_THEN
               (--`HOA (x':^Action)
                       (J AND (p pj V))
                       (SP V x' (J AND (p pj V)))`--)
               (fn thm=> ASSUME_TAC thm)
           THENL [REWRITE_TAC [SP_thm2], ALL_TAC]
           THEN REWRITE_TAC ([pj_pj] @ pj_transform)
           THEN UNDISCH_ALL_TAC
           THEN PRED_TAC_v3 [HOA_def] [],
           (*-- if the action is not runable --*)
           SUBGOAL_THEN
               (--`|== ((J:^Pred) AND (p pj V) 
                       IMP 
                       NOT (guard (x':^Action)))`--)
               ASSUME_TAC 
           THENL [ UNDISCH_ALL_TAC
                   THEN REWRITE_TAC [runSet_def]
                   THEN PROVE_TAC[],
                   ALL_TAC ]
           THEN REWRITE_TAC ([pj_pj] @ pj_transform)
           THEN SUBGOAL_THEN
                (--`HOA (x':^Action)
                     (NOT (guard x') AND
                        (J AND
                           ((J pj V) AND (p pj V) AND (guard x pj V))))
                     (J AND
                        ((J pj V) AND (p pj V) AND (guard x pj V)))`--)
                (fn thm=> ASSUME_TAC thm)
           THENL [ UNDISCH_ALL_TAC 
                   THEN REWRITE_TAC [disabledBy_def]
                   THEN REPEAT STRIP_TAC
                   THEN ASM_REWRITE_TAC[],
                   ALL_TAC ]
           THEN UNDISCH_ALL_TAC
           THEN PRED_TAC_v3 [HOA_def] []
          ],
        (*-- the enabling action --*)
        EXISTS_TAC (--`x:^Action`--)
        THEN CONJ_TAC
        THENL 
         [ UNDISCH_ALL_TAC THEN RW_TAC bool_ss [runSet_def],
           ALL_TAC ]
        THEN SUBGOAL_THEN
             (--`HOA (x:^Action)
                     (J AND (p pj V))
                     (SP V x (J AND (p pj V)))`--)
             (fn thm=> ASSUME_TAC thm)
        THENL [REWRITE_TAC [SP_thm2], ALL_TAC]
        THEN REWRITE_TAC ([pj_pj] @ pj_transform)
        THEN UNDISCH_ALL_TAC
        THEN PRED_TAC_v3 [HOA_def] []
        THEN REWRITE_TAC pj_transform
     ]
  ) ;


val N_LEADSTO_def = Define
  `(N_LEADSTO (Pr:^Umachine) guard V J 0 p q 
       = 
       Pvalid V J (p IMP q)) 

    /\

   (N_LEADSTO Pr guard V J (SUC n) p q
       =
       Pvalid V J (p AND NOT q
                   IMP
                   ProgressSpace Pr guard V J (p AND NOT q))
       /\
       N_LEADSTO Pr guard V J n
                 (NEXT Pr guard V J (p AND NOT q))
                 q
   )` ;



val SP_TO_LEADSTO_thm1 = store_thm
  ("SP_TO_LEADSTO_thm1",
 --`Pr a /\
    STABLE Pr J /\
    (!a. Pr a ==> disabledBy a (NOT (guard a))) /\
    N_LEADSTO Pr guard V J n p q
    ==>
    LEADSTO (Pr:^Umachine) V J p q`--,

    STRIP_TAC
    THEN UNDISCH_TAC 
         (--`N_LEADSTO (Pr:^Umachine) guard V J n p q`--)
    THEN SPEC_TAC (--`q:^Pred`--,--`q:^Pred`--)
    THEN SPEC_TAC (--`p:^Pred`--,--`p:^Pred`--)
    THEN Induct_on `n`
    THENL
    [ (*-- n=0 --*)
      REWRITE_TAC [N_LEADSTO_def]
      THEN PROVE_TAC [LEADSTO_IMP_LIFT],
      (*-- n=n+1 --*)
      REWRITE_TAC [N_LEADSTO_def]
      THEN REPEAT STRIP_TAC
      THEN SUBST1_TAC 
           (prove(--`(q:^Pred) = q OR q`--, PRED_TAC_v3 [] [])) 
      THEN SUBST1_TAC 
           (prove(--`(p:^Pred) = (p AND NOT q) OR (p AND q)`--,
                  PRED_TAC_v3 [] []))  
      THEN MATCH_MP_TAC LEADSTO_DISJ2 
      THEN CONJ_TAC
      THENL
      [ (*-- part of p in which q doesn't hold --*)
         PROVE_TAC [NEXT_LEADSTO_thm, LEADSTO_TRANS],
        (*-- part of p in which q already holds --*)
        MATCH_MP_TAC LEADSTO_IMP_LIFT
        THEN PRED_TAC_v3 pj_transform []   
      ]
   ]  
  );      

 
val _ = export_theory() ;



