(*-----------------------------------------------------------------------------------
  
  Our general progress operator (variant of leads-to) is defined as the
  smalles transitive and disjunctive closure of a 'single' action 
  progress. Many property of progress turns out to be derivable only
  from the fact that it is a disjunctive and transitive closure.

  This theory will contains such general properties of progress. The
  properties will be expressed in terms of arbitrary disjunctive and
  transitive closure.

   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.uu.nl

   Last update Apr 99
-----------------------------------------------------------------------------------*)           

open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

val _ = new_theory "utdc" ;

(*
load "bossLib" ;
load "restr_binderTheory" ;
load "uwfTheory" ;
load "btacs" ;
*)

open bossLib ;
open restr_binderTheory ;
open upredTheory ;
open uwfTheory ;
open urelTheory ;
open btacs ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;

(*-----------------------------------------------------------------------------------

  PART I: transitive and left disjunctive relations.

  Some basic definitions:

     U SUBREL V   : U is a sub-relation of V (defined in 'relation')
     REFL U       : U is a reflexive relation (defined in 'relation')
     TRANS U      : U is a transitive relation (defined in 'relation')
     Ri_TR U V    : V is the right transitive extension of U
     LDISJ W U    : U is a left disjunctive relation, wrt W
     DISJ W U     : U is a left and right over W disjunctive relation
     U rINTER V   : the intersection of U and V (defined in 'relation')


------------------------------------------------------------------------------------*)

val LDISJ_DEF = Define
   `LDISJ U =  !W (y:'a->bool).
                  (?x. W x) /\
                  (!x:'a->bool::W. U x y) ==> U (??x::W. x) y` ;

(*-- interestingly, I really need the non-emptiness assumption of W
     in the definition of LDISJ, or else I won't be able to prove
     that LEADSTO Pr Af V J p q implies STABLE Pr Af J.
-- *)

val DISJ_DEF = Define
   `DISJ W U =  !(f:'index->('a->bool)) 
                 (g:'index->('a->bool)).
                 (!i::W. U (f i) (g i))
                 ==> 
                 U (??i::W. f i) (??i::W. g i)` ;

val Ri_TR_DEF  = Define
    `Ri_TR U V = !(x:'a) y (z:'a). U x y /\ V y z ==> V x z` ; 

(*-------------------------------------------------------------------------------
   Some properties:
        - rINTER produces a lower bound
        - SUBREL is transitive
-------------------------------------------------------------------------------*)
        
val Pred = ty_antiq (Type `:'a->bool`) ;
val RelT = ty_antiq (Type `:^Pred->^Pred->bool`) ;

val rINTER_LB = store_thm 
  ("rINTER_LB",
   Term `!(U:^RelT) V. (U rINTER V) SUBREL V`,
     REWRITE_TAC [SUBREL_def, rINTER_def] THEN BETA_TAC
     THEN PROVE_TAC[]) ;

val SUBREL_TRANS = store_thm 
  ("SUBREL_TRANS",
   Term `TRANS ($SUBREL:^RelT->^RelT->bool)`,
    REWRITE_TAC [TRANS_def, SUBREL_def]
    THEN PROVE_TAC[]) ;  
    
(*-----------------------------------------------------------------------------
  Below are some simple disjuctivity properties.

    GEN_LDISJ    : a more generalized formulation of left disjuctivity
    GEN_RDISJ    : right disjuctivity
    SIMPLE_DISJ  : simple disjunctivity
    CANCEL       : calcelation law 

-----------------------------------------------------------------------------*)
 
fun APP_TRANS_TAC U =
    MATCH_MP_TAC (REWRITE_RULE [TRANS_def] 
                 (ASSUME (Term `TRANS ^U`))) ;

fun APP_LDISJ_TAC U = 
    MATCH_MP_TAC (REWRITE_RULE [LDISJ_DEF] 
                 (ASSUME (Term `LDISJ ^U`))) ;

val lemma1 = prove
  (Term `(??i:'index::W. f i) = (??x:^Pred::(\p. ?i. W i /\ (p = f i)). x)`,
   REWRITE_TAC [RES_qOR_def]
   THEN CONV_TAC (FUN_EQ_CONV)
   THEN BETA_TAC THEN PROVE_TAC[]) ;


val lemma2 = 
    let 
    val lemma = ((SPEC (Term `\p:^Pred. ?i:'index. W i /\ (p = f i)`))
              o (REWRITE_RULE [LDISJ_DEF])
              o ASSUME) (Term `LDISJ (U:^RelT)`) 
    in
    BETA_RULE(RESTRICT_SHIFT lemma) 
    end ;

val LDISJ_GEN = store_thm 
   ("LDISJ_GEN",
    Term `!(U:^RelT) W f q.
          LDISJ U /\ 
          (?i. W i) /\
          (!i::W. U (f i) q)
          ==>
          U (??i::W. f i) q`,
    
    RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THEN REWRITE_TAC[lemma1]
    THEN MATCH_MP_TAC lemma2 
    THEN UNDISCH_ALL_TAC THEN PROVE_TAC[]) ;


val lemma = (UNDISCH o prove)
  (Term `W i ==> (p = (??i:'index::W. (\i. (p:^Pred)) i))`,
    REWRITE_TAC [RES_qOR_def]
    THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val GEN_RDISJ = store_thm 
  ("GEN_RDISJ",
    Term `!(U:^RelT) W f p.
         DISJ W U /\ (?i. W i) /\ (!(i:'index)::W. U p (f i))
         ==>
         U p (??i::W. f i)`,
     REWRITE_TAC [DISJ_DEF]
     THEN RESTRICT_DEF_TAC
     THEN REPEAT STRIP_TAC
     THEN SUBST_TAC [lemma]
     THEN FIRST_ASSUM MATCH_MP_TAC
     THEN BETA_TAC THEN ASM_REWRITE_TAC[]) ;

val GEN_LDISJ = store_thm 
  ("GEN_LDISJ",
    Term `!(U:^RelT) W f q.
     DISJ W U /\ (?i. W i) /\ (!i:'index::W. U (f i) q)
     ==>
     U (??i::W. f i) q`,
     REWRITE_TAC [DISJ_DEF]
     THEN REPEAT STRIP_TAC
     THEN SUBST_TAC [SPEC (Term `q:^Pred`) (GEN (Term `p:^Pred`) lemma)]
     THEN FIRST_ASSUM MATCH_MP_TAC
     THEN BETA_TAC THEN ASM_REWRITE_TAC[]) ;
            
val lemma2 = prove 
   (Term `(p:^Pred) OR q = (??i::TT. (\i. i => p | q) i)`,
    REWRITE_TAC [OR_def, TT_def, RES_qOR_def]
    THEN CONV_TAC FUN_EQ_CONV
    THEN BETA_TAC 
    THEN CONV_TAC (DEPTH_CONV COND_CONV)
    THEN GEN_TAC THEN EQ_TAC
    THEN STRIP_TAC
    THENL [ EXISTS_TAC (Term `T`) THEN ASM_REWRITE_TAC[] ,
            EXISTS_TAC (Term `F`) THEN ASM_REWRITE_TAC[] ,
            ASM_CASES_TAC (Term `i:bool`)
            THEN UNDISCH_TAC (Term `(i => (p:^Pred) | q)x`)
            THEN ASM_REWRITE_TAC[]
            THEN PROVE_TAC[] ]) ;

val lemma3 = SPECL [Term `r:^Pred`, Term `p:^Pred`] (GEN_ALL lemma2) ;
val lemma4 = SPECL [Term `s:^Pred`, Term `q:^Pred`] (GEN_ALL lemma2) ;


val SIMPLE_DISJ = store_thm 
  ("SIMPLE_DISJ",
    Term `!(U:^RelT) p q r s.
             DISJ (TT:bool->bool) U /\ U p q /\ U r s
             ==>
             U (p OR r) (q OR s)`,
     REWRITE_TAC [DISJ_DEF]
     THEN RESTRICT_DEF_TAC 
     THEN REPEAT STRIP_TAC
     THEN SUBST_TAC [lemma3,lemma4]
     THEN FIRST_ASSUM MATCH_MP_TAC 
     THEN REWRITE_TAC [TT_def] THEN BETA_TAC
     THEN GEN_TAC THEN ASM_CASES_TAC (Term `x:bool`)
     THEN ASM_REWRITE_TAC[]) ;
    
val CANCEL = store_thm 
  ("CANCEL",
   Term `!(U:^RelT) p q r s.    
         TRANS U  /\ DISJ (TT:bool->bool) U /\
         U q q /\ U p (q OR r) /\ U r s
         ==>
         U p (q OR s)`,
     REPEAT STRIP_TAC
     THEN APP_TRANS_TAC (Term `U:^RelT`)
     THEN EXISTS_TAC (Term `(q:^Pred) OR r`)
     THEN ASM_REWRITE_TAC[]
     THEN MATCH_MP_TAC SIMPLE_DISJ
     THEN ASM_REWRITE_TAC[]) ;

(*--------------------------------------------------------------------------------
    The principle of reachability by bounded decrement:

    Given a transitive, left disjunctive, and IMP-containing relation -->:AxA,
          a well-founded relation <<:DxD
          a metric function M:A->D

    the principle says (informally!):

             !m. (p /\ M=m) --> ((p /\ M<m) \/ q)
           ---------------------------------------
                          p --> q

---------------------------------------------------------------------------------*)

val MRel   = ty_antiq (Type `:'m->'m->bool`) ;
val Metric = ty_antiq (Type `:'a->'m`) ;

val lemma = prove
    (Term `((??x:'m::(\x. LESS x (y:'m)). p AND (\s:'a. M s = x)) OR q) =
      ((p AND (\s. LESS (M s) y)) OR q)`,
     REWRITE_TAC [RES_qOR_def, AND_def, OR_def]
     THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
     THEN BETA_TAC THEN PROVE_TAC[]) ;

val lemma1 = prove
    (Term `~(?i:'index. W i) ==> ((??i::W. (P i):^Pred) = FF)`,
     (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
     THEN REWRITE_TAC [RES_qOR_def, FF_def]
     THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
     THEN BETA_TAC THEN PROVE_TAC[]) ;


val BOUNDED_REACH_lemma  = prove
  (Term `!(LESS:^MRel) (M:^Metric) (U:^RelT) p q.
        ADMIT_WF_INDUCTION LESS /\        
        TRANS U /\ LDISJ U /\ DISJ (TT:bool->bool) U /\ 
        U q q /\
        (!m. U (p AND (\s. M s = m)) ((p AND (\s. LESS (M s) m)) OR q))
        ==>
        (!m. U (p AND (\s. M s = m)) q)`,
    REWRITE_TAC [ADMIT_WF_INDUCTION_def]    
    THEN REPEAT GEN_TAC THEN STRIP_TAC
    THEN ((CONV_TAC o RAND_CONV o ABS_CONV) (MK_ABS_CONV (Term `m:'m`)))
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN APP_TRANS_TAC (Term `U:^RelT`)
    THEN EXISTS_TAC 
         (Term `(??x:'m::(\x. LESS x (y:'m)). (p AND (\s:'a. M s = x))) OR q`)
    THEN CONJ_TAC
         (*--- Split the two transitive cases --*)
    THENL 
    [ (*- (1) -*)
      ASM_REWRITE_TAC [lemma] ,
      (*- (2) -*)
      ASM_CASES_TAC (Term `?x. (\x. (LESS:^MRel) x y)x`)
      THENL
      [ (*- 2.1 y is not bottom -*)
        (CONV_TAC o RAND_CONV o ONCE_REWRITE_CONV) [SYM(SPEC_ALL pOR_REFL)]
        THEN MATCH_MP_TAC SIMPLE_DISJ
        THEN ASM_REWRITE_TAC[] 
        THEN (CONV_TAC o ONCE_DEPTH_CONV o RAND_CONV o ABS_CONV)
             (MK_ABS_CONV (Term `x:'m`))
        THEN MATCH_MP_TAC LDISJ_GEN
        THEN ASM_REWRITE_TAC[]
        THEN RESTRICT_DEF_TAC
        THEN ASM_REWRITE_TAC[] ,
        (*- 2.2 y is bottom -*)
        (CONV_TAC o ONCE_DEPTH_CONV o RAND_CONV o ABS_CONV)
             (MK_ABS_CONV (Term `x:'m`))
        THEN IMP_RES_TAC lemma1
        THEN ASM_REWRITE_TAC[]
        THEN ASM_REWRITE_TAC [pOR_UNIT]]
    ])  ;


val lemma2 = prove
   (Term `p = (??m:'m::TT. p AND (\s:'a. M s = m))`,
     REWRITE_TAC [RES_qOR_def, AND_def, TT_def]
     THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
     THEN BETA_TAC THEN PROVE_TAC[]) ;


val BOUNDED_REACH = store_thm 
  ("BOUNDED_REACH",
   Term `!(LESS:^MRel) (M:^Metric) (U:^RelT) p q.
        ADMIT_WF_INDUCTION LESS /\        
        TRANS U /\ LDISJ U /\ DISJ (TT:bool->bool) U /\ 
        U q q /\
        (!m. U (p AND (\s. M s = m)) ((p AND (\s. LESS (M s) m)) OR q))
        ==>
        U p q`,
   REPEAT STRIP_TAC
   THEN SUBST1_TAC lemma2
   THEN ((CONV_TAC o ONCE_DEPTH_CONV o RAND_CONV o ABS_CONV)
         (MK_ABS_CONV (Term `m:'m`)))
   THEN MATCH_MP_TAC LDISJ_GEN
   THEN ASM_REWRITE_TAC[] 
   THEN RESTRICT_DEF_TAC THEN REWRITE_TAC [TT_def]
   THEN IMP_RES_TAC BOUNDED_REACH_lemma) ;

(*--------------------------------------------------------------------------------
    A special case of the reachability by bounded decrement principle is the
    following: 

             !m. (~p /\ M=m) --> M<m
           ----------------------------
                    true --> p

---------------------------------------------------------------------------------*)

val predREDUCE_TAC =
   REWRITE_TAC [AND_def, OR_def, IMP_def, NOT_def, TT_def, FF_def, EQUAL_def,
                RES_qAND_def, RES_qOR_def, pSEQ_DEF]
   THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
   THEN BETA_TAC ;

val predPROVE_TAC = predREDUCE_TAC THEN PROVE_TAC[]
                

val BOUNDED_ALWAYS_REACH = store_thm (
   "BOUNDED_ALWAYS_REACH",
   Term `!(LESS:^MRel) (M:^Metric) U p.
        ADMIT_WF_INDUCTION LESS /\
        TRANS U /\ LDISJ U /\ DISJ (TT:bool->bool) U /\ 
        U p p /\
        (!m. U ((NOT p) AND (\s. M s = m)) ((\s. LESS (M s) m) OR p))
        ==> 
        U TT p`,
    REPEAT STRIP_TAC
    THEN SUBST1_TAC (SYM(ISPEC (Term `p:^Pred`) pOR_REFL))
    THEN SUBST1_TAC (prove(Term `TT=((p:^Pred) OR (NOT p))`, predPROVE_TAC))
    THEN MATCH_MP_TAC SIMPLE_DISJ
    THEN ASM_REWRITE_TAC[] 
    THEN MATCH_MP_TAC BOUNDED_REACH
    THEN EXISTS_TAC (Term `LESS:^MRel`)
    THEN EXISTS_TAC (Term `M:^Metric`)
    THEN ASM_REWRITE_TAC []
    THEN ASM_REWRITE_TAC 
         [ prove(Term `(((NOT p) AND q) OR (p:^Pred)) = (q OR p)`,
                 predPROVE_TAC)]) ;

(*--------------------------------------------------------------------------------

   PART II : least transitive and disjunctive closures.

   Some basic definitions:

      TC U     : the least transitive closure of U
      RTC U    : the least reflexive and transitive closure of U
      TDC U    : the least transitive and left disjuctive closure of U
      Ri_TDC U : the least right transitive and left disjuctive closure of U


   As will be shown latter, TDC = Ri_TDC. Ri_TDC provides a slightly 
   stronger induction hypothesis, which may be useful in some cases.

---------------------------------------------------------------------------------*)

val TransC_DEF = Define
   `TransC (U:'a->'a->bool) x y = !X. U SUBREL X /\ TRANS X ==> X x y` ;

val RTC_DEF = Define
   `RTC (U:'a->'a->bool) x y = !X. U SUBREL X /\ REFL X /\ TRANS X ==> X x y` ;

val TDC_DEF = Define
    `TDC U (x:^Pred) y = !X. U SUBREL X /\ TRANS X /\ LDISJ X ==> X x y` ;

val Ri_TDC_DEF = Define
   `Ri_TDC U (x:^Pred) y = !X. U SUBREL X /\ Ri_TR U X /\ LDISJ X ==> X x y` ;

(*-----------------------------------------------------------------------------------
       TC U, RTC U, and TDC U all include U and transitive
       RTC U is reflexive and TDC U is left disjunctive
-----------------------------------------------------------------------------------*)

val TransC_LIFT = store_thm 
  ("TransC_LIFT",
    Term `!(U:'a->'a->bool). U SUBREL (TransC U)`,
    REWRITE_TAC [SUBREL_def, TransC_DEF]
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;

val RTC_LIFT = store_thm 
  ("RTC_LIFT",
    Term `!(U:'a->'a->bool). U SUBREL (RTC U)`,
    REWRITE_TAC [SUBREL_def, RTC_DEF]
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;

val TDC_LIFT = store_thm 
  ("TDC_LIFT",
    Term `!(U:^RelT). U SUBREL (TDC U)`,
    REWRITE_TAC [SUBREL_def, TDC_DEF]
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;

val TransC_TRANS = store_thm 
  ("TransC_TRANS",
    Term `!(U:'a->'a->bool). TRANS (TransC U)`,
    REWRITE_TAC [TRANS_def, TransC_DEF]
    THEN REPEAT STRIP_TAC THEN REPEAT RES_TAC) ;

val RTC_TRANS = store_thm 
  ("RTC_TRANS",
    Term `!(U:'a->'a->bool). TRANS (RTC U)`,
    REWRITE_TAC [TRANS_def, RTC_DEF]
    THEN REPEAT STRIP_TAC THEN REPEAT RES_TAC) ;

val TDC_TRANS = store_thm 
  ("TDC_TRANS",
    Term `!(U:^RelT). TRANS (TDC U)`,
    REWRITE_TAC [TRANS_def, TDC_DEF]
    THEN REPEAT STRIP_TAC THEN REPEAT RES_TAC) ;

val TDC_Ri_TRANS = store_thm 
  ("TDC_Ri_TRANS",
    Term `!(U:^RelT). Ri_TR U (TDC U)`,
    REWRITE_TAC [Ri_TR_DEF]
    THEN REPEAT STRIP_TAC 
    THEN IMP_RES_TAC (REWRITE_RULE [SUBREL_def] TDC_LIFT)
    THEN IMP_RES_TAC (REWRITE_RULE [TRANS_def] TDC_TRANS)) ;

val RTC_REFL = store_thm 
  ("RTC_REFL",
    Term `!(U:'a->'a->bool). REFL (RTC U)`,
    REWRITE_TAC [REFL_def, RTC_DEF]
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]) ;

val TDC_LDISJ = store_thm 
  ("TDC_LDISJ",
    Term `!(U:^RelT). LDISJ (TDC U)`,
    REWRITE_TAC [LDISJ_DEF, TDC_DEF]
    THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN (Term `!x. W x ==> (X:^RelT) x y`)
         (fn thm=> ASSUME_TAC thm) 
    THENL [ REPEAT STRIP_TAC THEN RES_TAC , RES_TAC]) ;

(*-------------------------------------------------------------------------------
                           RTC includes TC
-------------------------------------------------------------------------------*)

val TransC_IMP_RTC = store_thm 
  ("TransC_IMP_RTC",
    Term `!(U:'a->'a->bool). (TransC U) SUBREL (RTC U)`,
    REWRITE_TAC [TransC_DEF, SUBREL_def]
    THEN REPEAT STRIP_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC [REWRITE_RULE [SUBREL_def] RTC_LIFT, RTC_TRANS]) ;
 
(*-------------------------------------------------------------------------------
   an induction princile for TDC:
      
   To prove that a relation X includes TDC U it suffices to show that
      
       (1) X includes U
       (2) TRANS X
       (3) DISJ X
      
-------------------------------------------------------------------------------*)

val TDC_INDUCT1 = store_thm 
  ("TDC_INDUCT1",
    Term `!(U:^RelT) X. U SUBREL X /\ TRANS X /\ LDISJ X ==> (TDC U) SUBREL X`,
    REWRITE_TAC [SUBREL_def, TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;

fun MK_2ABS_CONV p q (trm:term) =
    SYM (((ONCE_DEPTH_CONV BETA_CONV) 
          THENC BETA_CONV) (Term `(\^p ^q. ^trm) ^p ^q`)) ;

val TDC_INDUCT2 = store_thm 
  ("TDC_INDUCT2",
   Term  `!(U:^RelT) X.  
        U SUBREL ((TDC U) rINTER X) /\ 
        TRANS ((TDC U) rINTER X)    /\ 
        LDISJ ((TDC U) rINTER X) 
        ==> (TDC U) SUBREL X`,
    REWRITE_TAC [SUBREL_def, TRANS_def, LDISJ_DEF, rINTER_def]
    THEN BETA_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC
    THEN REPEAT GEN_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [TRANS_def, SUBREL_def] SUBREL_TRANS)
    THEN EXISTS_TAC (Term `(TDC (U:^RelT)) rINTER X`)
    THEN REWRITE_TAC [rINTER_def] THEN BETA_TAC
    THEN CONJ_TAC
    THENL [ ALL_TAC , PROVE_TAC[]]
    THEN REPEAT GEN_TAC 
    THEN CONV_TAC (RAND_CONV (MK_2ABS_CONV (Term `x':^Pred`) (Term `y':^Pred`)))
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL_def] TDC_INDUCT1)
    THEN REWRITE_TAC [TRANS_def, LDISJ_DEF]
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;


(*-----------------------------------------------------------------------------
   Some monotonicity properties of TDC:                 

     (1) if V includes U, then TDC V includes U 
     (2) TDC is monotonic wrt SUBREL  

-----------------------------------------------------------------------------*)

val TDC_SUBREL = store_thm 
  ("TDC_SUBREL",
   Term `!(U:^RelT) V. (U SUBREL V) ==> (U SUBREL (TDC V))`,
    REWRITE_TAC [SUBREL_def]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL_def] TDC_LIFT)
    THEN RES_TAC) ;
    
val TDC_MONO = store_thm 
  ("TDC_MONO",
    Term `!(U:^RelT) V. (U SUBREL V) ==> (TDC U) SUBREL (TDC V)`,
    REPEAT STRIP_TAC 
    THEN MATCH_MP_TAC TDC_INDUCT1
    THEN REWRITE_TAC [TDC_TRANS, TDC_LDISJ]
    THEN MATCH_MP_TAC TDC_SUBREL
    THEN ASM_REWRITE_TAC[]) ;

(*------------------------------------------------------------------------------
   Ri_TDC U is a closure of U, right transitive, and disjunctive
------------------------------------------------------------------------------*)

val Ri_TDC_LIFT = store_thm 
  ("Ri_TDC_LIFT",
    Term `!(U:^RelT). U SUBREL (Ri_TDC U)`,
    REWRITE_TAC [SUBREL_def, Ri_TDC_DEF]
    THEN REPEAT STRIP_TAC THEN RES_TAC) ;

val Ri_TDC_Ri_TRANS = store_thm 
  ("Ri_TDC_Ri_TRANS",
   Term  `!(U:^RelT). Ri_TR U (Ri_TDC U)`,
    REWRITE_TAC [Ri_TR_DEF, Ri_TDC_DEF]
    THEN REPEAT STRIP_TAC THEN REPEAT RES_TAC) ;

val Ri_TDC_LDISJ = store_thm 
  ("Ri_TDC_LDISJ",
   Term  `!(U:^RelT). LDISJ (Ri_TDC U)`,
    REWRITE_TAC [LDISJ_DEF, Ri_TDC_DEF]
    THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC 
    THEN SUBGOAL_THEN (Term `!x. W x ==> (X:^RelT) x y`)
                      ASSUME_TAC 
    THENL [ REPEAT STRIP_TAC THEN RES_TAC , RES_TAC]) ;

(*-------------------------------------------------------------------------------
                  an induction principle for Ri_TDC
-------------------------------------------------------------------------------*)
 
val Ri_TDC_INDUCT1 = store_thm 
  ("Ri_TDC_INDUCT1",
   Term `!(U:^RelT) X.  U SUBREL X  /\  Ri_TR U X  /\  LDISJ X 
       ==> (Ri_TDC U) SUBREL X`,
    REWRITE_TAC [SUBREL_def, Ri_TDC_DEF]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;

(*-------------------------------------------------------------------------------
                 another induction principle for Ri_TDC
-------------------------------------------------------------------------------*)

val Ri_TDC_INDUCT2 = store_thm 
  ("Ri_TDC_INDUCT2",
   Term `!(U:^RelT) X. 
       U SUBREL X  /\ 
       (!x y z. U x y /\ X y z /\ Ri_TDC U y z ==> X x z)  /\
       (!W y. (?x. W x) /\
              (!x::W. X x y /\ Ri_TDC U x y) ==> X (??x::W.x) y)
       ==>
       (Ri_TDC U) SUBREL X`,
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [TRANS_def] SUBREL_TRANS)
    THEN EXISTS_TAC (Term `(Ri_TDC U) rINTER (X:^RelT)`)
    THEN REWRITE_TAC[rINTER_LB] 
    THEN MATCH_MP_TAC Ri_TDC_INDUCT1 
    THEN UNDISCH_ALL_TAC
    THEN REWRITE_TAC [SUBREL_def, rINTER_def, Ri_TR_DEF, LDISJ_DEF]
    THEN BETA_TAC THEN RESTRICT_DEF_TAC
    THEN REPEAT STRIP_TAC
    THENL [ IMP_RES_TAC (REWRITE_RULE [SUBREL_def] Ri_TDC_LIFT) ,
            RES_TAC ,
            IMP_RES_TAC (REWRITE_RULE [Ri_TR_DEF] Ri_TDC_Ri_TRANS) ,
            RES_TAC ,
            MATCH_MP_TAC (REWRITE_RULE [LDISJ_DEF] Ri_TDC_LDISJ)          
            THEN RESTRICT_DEF_TAC
            THEN REPEAT STRIP_TAC THEN PROVE_TAC[] ,
            PROVE_TAC[] ]) ;


(*------------------------------------------------------------------------------
   Ri_TDC U is also transitive (so, not only right-transitive wrt U)
------------------------------------------------------------------------------*)

val lemma = prove
  (Term  `((X (x:'a) (y:'a)) /\ X y z ==> X x z) =
          (X x y ==> ((\x y. X y z ==> X x z) x y))`,
   PROVE_TAC[]) ;
            

val Ri_TDC_TRANS = store_thm 
  ("Ri_TDC_TRANS",
    Term `!(U:^RelT). TRANS (Ri_TDC U)`,
    REWRITE_TAC [TRANS_def, lemma]
    THEN REPEAT GEN_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE [SUBREL_def] Ri_TDC_INDUCT1)
    THEN BETA_TAC THEN REPEAT CONJ_TAC
    THENL [ REPEAT STRIP_TAC
            THEN IMP_RES_TAC (REWRITE_RULE [Ri_TR_DEF] Ri_TDC_Ri_TRANS) ,
            REWRITE_TAC [Ri_TR_DEF]
            THEN BETA_TAC THEN REPEAT STRIP_TAC
            THEN RES_TAC
            THEN IMP_RES_TAC (REWRITE_RULE [Ri_TR_DEF] Ri_TDC_Ri_TRANS) ,
            REWRITE_TAC [LDISJ_DEF]
            THEN BETA_TAC THEN RESTRICT_DEF_TAC
            THEN REPEAT STRIP_TAC
            THEN SUBGOAL_THEN (Term `!x::W. Ri_TDC (U:^RelT) x z`)
                 ASSUME_TAC 
            THENL [ RESTRICT_DEF_TAC 
                    THEN REPEAT STRIP_TAC THEN RES_TAC ,
                    IMP_RES_TAC (REWRITE_RULE [LDISJ_DEF] Ri_TDC_LDISJ)]
          ]) ;

(*-------------------------------------------------------------------------------
                    TDC is equal to Ri_TDC
-------------------------------------------------------------------------------*)
 
val Ri_TDC_IMP_TDC = prove
  (Term `(Ri_TDC (U:^RelT)) SUBREL (TDC U)`,
   MATCH_MP_TAC Ri_TDC_INDUCT1
   THEN REWRITE_TAC [TDC_LIFT, TDC_Ri_TRANS, TDC_LDISJ]) ;

val TDC_IMP_Ri_TDC = prove
  (Term `(TDC (U:^RelT)) SUBREL (Ri_TDC U)`,
   MATCH_MP_TAC TDC_INDUCT1
   THEN REWRITE_TAC [Ri_TDC_LIFT, Ri_TDC_TRANS, Ri_TDC_LDISJ]) ;

val TDC_EQU_Ri_TDC = store_thm 
  ("TDC_EQU_Ri_TDC",
   Term `(TDC:^RelT -> ^RelT) = Ri_TDC`,
    REPEAT (CONV_TAC FUN_EQ_CONV THEN GEN_TAC)
    THEN EQ_TAC
    THEN REWRITE_TAC (map (fn thm=> REWRITE_RULE [SUBREL_def] thm)
                          [Ri_TDC_IMP_TDC , TDC_IMP_Ri_TDC])) ; 
    
(*-------------------------------------------------------------------------------------
                the double induction principle for Ri_TDC
-------------------------------------------------------------------------------------*)

val XX = Term `\(r:^Pred) (s:^Pred) (p:^Pred) (q:^Pred). 
                ri_TDC (U:^RelT) p q /\ X p q r s` ;
val Type_of_X  = ty_antiq (Type `:^Pred->^Pred->^Pred->^Pred->bool`) ;

val DOUBLE_INDUCT_ante =
    Term
    `(!p q r s. (X:^Type_of_X) p q r s = X r s p q) /\
     (!p q r s. U p q ==> X p q r s) /\
     (!W q r s. 
         (?p. W p) /\
         (!p :: W. Ri_TDC U p q /\ X p q r s) 
         ==> X (?? p :: W. p) q r s) /\   
     (!p q r u v w.  
         U p q /\ U u v /\ Ri_TDC U q r /\  Ri_TDC U v w /\
         X p r v w /\ X u w q r /\ X q r v w 
         ==> X p r u w)` ;

fun DISJ_TAC x =
    FIRST_ASSUM (fn thm=> MATCH_MP_TAC (SPEC (Term `W:^Pred->bool`) thm))
    THEN CONJ_TAC
    THENL [ EXISTS_TAC x THEN ASM_REWRITE_TAC[] , ALL_TAC ] 
    THEN RESTRICT_DEF_TAC THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC RESTRICT_SHIFT THEN RES_TAC ;


val DOUBLE_INDUCT_thm = prove(
     Term `^DOUBLE_INDUCT_ante
     ==>
     (Ri_TDC (U:^RelT)) SUBREL 
         ((\p q. (Ri_TDC U) SUBREL (\r s. X p q r s)))`,
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC Ri_TDC_INDUCT2
    THEN REWRITE_TAC [Ri_TR_DEF, LDISJ_DEF, rINTER_def] 
    THEN BETA_TAC
    THEN REPEAT CONJ_TAC THEN REPEAT STRIP_TAC
    THENL 
    [ (*- (1) Lift Case --------------*)
      REWRITE_TAC [SUBREL_def, rINTER_def]
      THEN BETA_TAC THEN REPEAT STRIP_TAC 
      THEN XRULE_ASSUM_TAC (SPECL [Term `x:^Pred`, Term `y:^Pred`, 
                                   Term `x':^Pred`,Term `y':^Pred`])
      THEN RES_TAC ,
      (*-- (2) transitivity case -----*)
      MATCH_MP_TAC Ri_TDC_INDUCT2
      THEN UNDISCH_ALL_TAC
      THEN REWRITE_TAC [SUBREL_def, rINTER_def, DISJ_DEF, Ri_TR_DEF] 
      THEN BETA_TAC 
      THEN REPEAT STRIP_TAC
      THENL 
      [ (*- (2.1) Lift Case --------------*)
        EVERY_ASSUM (fn thm=>ONCE_REWRITE_TAC[thm])
        THEN MATCH_MP_TAC (ASSUME
            (Term 
             `!(p:^Pred) (q:^Pred) (r:^Pred) (s:^Pred). 
              U p q ==> X p q r s`))
        THEN FIRST_ASSUM ACCEPT_TAC ,
        (*-- (2.2) transitivity case -----*)
        IMP_RES_TAC (REWRITE_RULE [Ri_TR_DEF] Ri_TDC_Ri_TRANS)
        THEN UNDISCH_ALL_TAC THEN PROVE_TAC[] ,
        (*-- (2.3) disjunctive case -----*)
        EVERY_ASSUM (fn thm=> ONCE_REWRITE_TAC [thm])
        THEN MATCH_MP_TAC (ASSUME
            (Term 
             `!W q r s. (?p. W p) /\
                        (!p:^Pred::W. Ri_TDC U p q /\ X p q r s) 
                        ==> 
                        X (??p ::W. p) (q:^Pred) (r:^Pred) (s:^Pred)`))
        THEN UNDISCH_ALL_TAC THEN RESTRICT_DEF_TAC
        THEN PROVE_TAC[]  ] ,
      (*- (3) Disjunctivity case -------*)
      REWRITE_TAC [SUBREL_def]
      THEN REPEAT STRIP_TAC THEN BETA_TAC
      THEN MATCH_MP_TAC (ASSUME
            (Term 
             `!W q r s. (?p. W p) /\
                        (!p:^Pred::W. Ri_TDC U p q /\ X p q r s) 
                        ==> 
                        X (??p ::W. p) (q:^Pred) (r:^Pred) (s:^Pred)`))
      THEN UNDISCH_ALL_TAC 
      THEN REWRITE_TAC [SUBREL_def]
      THEN RESTRICT_DEF_TAC
      THEN PROVE_TAC[] 
   ] ) ;


val Ri_TDC_DOUBLE_INDUCT = store_thm 
  ("Ri_TDC_DOUBLE_INDUCT",
    Term `!(U:^RelT) X. 
     ^DOUBLE_INDUCT_ante
     ==>
     (!p q r s. Ri_TDC U p q /\ Ri_TDC U r s ==> X p q r s)`,
     REPEAT STRIP_TAC
     THEN IMP_RES_TAC (BETA_RULE (REWRITE_RULE [SUBREL_def] 
                                  DOUBLE_INDUCT_thm))) ;

val _ = export_theory () ;

(*-----------------------------------------------------------------------------------
  
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.ruu.nl

-----------------------------------------------------------------------------------*)           
