(* --------------------------------------------------------------------------

  This library contains various basic theorem about lifted predicate calculus.
  A predicate over type A is a function  of type A -> bool. Predicate is used
  extensively in programming to describe possible states that a program may
  visit during its execution.

  Predicate operators are boolean operator lifted to the predicate level.
  For example:
  
      p AND q = (\s. p s /\ q s)

  Various laws of the boolean predicate calculus also holds for the lifted
  predicate calculus. 

  A predicate p over A holds everywhere if for all x element of A, p s holds.
  We denote it:

      |== p

  (In some books it is denoted [p])

  author: Wishnu Prasetya
          Dept. of Comp. Science
          Utrecht University
          (wishnu@cs.uu.nl)
  
  last update: feb 1999

----------------------------------------------------------------------------*)

open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;
  
val _ = new_theory "upred" ;

(*
load "bossLib" ; 
load "restr_binderTheory" ;
load "setTheory" ; *)


open bossLib ;
open restr_binderTheory ;
open setTheory ;

(*----------------------------------------------------------------------------
                       some help tactics
----------------------------------------------------------------------------*)


local fun EXT_SPEC_TAC (asml,g) = REWRITE_TAC [FUN_EQ_CONV g] (asml,g)
in
val EXT_TAC  =
    REPEAT GEN_TAC 
    THEN EXT_SPEC_TAC 
    THEN BETA_TAC 
    THEN GEN_TAC 
end ;


fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;


(*---------------------------------------------------------------------------
  Definitions supported:

      TT s                =  T
      FF s                =  F
      (NOT p) s           = ~p s
      (p AND q) s         =  p s /\ q s
      (p OR q) s          = p s \/ q s
      (p IMP q) s         = p s ==> q s
      (p EQUAL q) s       = (p s = q s)
      (!!i::P. Q i) s     = (!i::P. Q i s)
      (??i::P. Q i) s     = (?i::P. Q i s)
      |== p               = (!s. p s)

---------------------------------------------------------------------------*)

val TT_DEF = Define `(TT:'a->bool) = \s:'a.T` ;

val FF_DEF = Define `(FF:'a->bool) = \s:'a.F` ;

val pNOT_DEF = Define `NOT (p:'a->bool) = \s:'a. ~p s` ;

val pAND_DEF = Define `AND p q = \s:'a. p s /\ q s` ;
val _ = set_fixity "AND" (Infix 600) ;

val pOR_DEF = Define `OR p q = \s:'a. p s \/ q s` ;
val _ = set_fixity "OR" (Infix 580) ;

val pIMP_DEF = Define `IMP p q = \s:'a. p s ==> q s` ;
val _ = set_fixity "IMP" (Infix 560) ;

val EQUAL_DEF = Define `EQUAL p q = \s:'a. p s = q s` ;
val _ = set_fixity "EQUAL" (Infix 540) ;

val _ = new_binder {Name="!!", Ty=Type `:('b->bool)->'a->bool` }  ;
val _ = new_binder {Name="??", Ty=Type `:('b->bool)->'a->bool` } ;

val RES_qAND = Define `RES_qAND W P = (\s:'a. !i:'b. W i ==> P i s)` ;

val RES_qOR  = Define `RES_qOR W P  = (\s:'a. ?i:'b. W i /\  P i s)` ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;

val pSEQ_DEF = Define `|== p = !s:'a. p s` ;
val _ = set_MLname "|==_def" "pSEQ_DEF";

(*---------------------------------------------------------------------------
                                 pred_TAC 
---------------------------------------------------------------------------*)

val pred_TAC =
    REWRITE_TAC [FF_DEF, TT_DEF, pNOT_DEF, pAND_DEF, pOR_DEF, 
                 pIMP_DEF, EQUAL_DEF, RES_qAND, RES_qOR, pSEQ_DEF,
                 RES_FORALL, RES_EXISTS,
                 FUN_EQ_CONV (Term `(f:'a->'b)=g`)
                ]
    THEN PROVE_TAC [] ;

(*--------------------------------------------------------------------------*)

val pAND_SYM = store_thm
  ("pAND_SYM", 
   Term `!(p:'a->bool) q. (p AND q) = (q AND p)`, 
   pred_TAC) ;

val pAND_ASSOC = store_thm 
  ("pAND_ASSOC", 
   Term `!(p:'a->bool) q r. (p AND q) AND r = p AND (q AND r)`, 
   pred_TAC) ;

val pAND_UNIT = store_thm 
  ("pAND_UNIT", 
   Term `!p:'a->bool. (TT AND p) = p`, 
   pred_TAC) ;

val pOR_SYM = store_thm 
  ("pOR_SYM", 
   Term `!(p:'a->bool) q. (p OR q) = (q OR p)`, 
   pred_TAC) ;

val pOR_ASSOC = store_thm 
  ("pOR_ASSOC",
   Term `!(p:'a->bool) q r. (p OR q) OR r =  p OR (q OR r)`, 
   pred_TAC) ;

val pOR_UNIT = store_thm 
  ("pOR_UNIT", 
   Term `!p:'a->bool. (FF OR p) = p`, 
   pred_TAC) ;

val pAND_ZERO = store_thm 
  ("pAND_ZERO", 
   Term `!p:'a->bool. (FF AND p) = FF`, 
   pred_TAC) ;

val pAND_LOVER_OR = store_thm 
  ("pAND_LOVER_OR",
    Term `!(p:'a->bool) q r. p AND (q OR r) = (p AND q) OR (p AND r)`, 
    pred_TAC) ;

val pAND_ROVER_OR = store_thm 
  ("pAND_ROVER_OR",
    Term `!(p:'a->bool) q r. (p OR q) AND r = (p AND r) OR (q AND r)`, 
    pred_TAC) ;

val pOR_ZERO = store_thm 
  ("pOR_ZERO", 
    Term `!p:'a->bool. (TT OR p) = TT`, 
    pred_TAC) ;

val pOR_LOVER_AND = store_thm 
  ("pOR_LOVER_AND",
    Term `!(p:'a->bool) q r. p OR (q AND r) = (p OR q) AND (p OR r)`, 
    pred_TAC) ;

val pOR_ROVER_AND = store_thm 
  ("pOR_ROVER_AND",
    Term `!(p:'a->bool) q r. (p AND q) OR r = (p OR r) AND (q OR r)`, 
    pred_TAC) ;

val pAND_REFL = store_thm 
  ("pAND_REFL", 
    Term `!p:'a->bool. (p AND p) = p`, 
    pred_TAC) ;

val pOR_REFL = store_thm 
  ("pOR_REFL", 
    Term `!p:'a->bool. (p OR p) = p`, 
    pred_TAC) ;

val pAND_LOVER_AND = store_thm 
  ("pAND_LOVER_AND",
    Term `!(p:'a->bool) q r. p AND (q AND r) = (p AND q) AND (p AND r)`, 
    pred_TAC) ;

val pOR_LOVER_OR = store_thm 
  ("pOR_LOVER_OR",
    Term `!(p:'a->bool) q r. p OR (q OR r) = (p OR q) OR (p OR r)`, 
    pred_TAC) ;

val pAND_ROVER_AND = store_thm 
  ("pAND_ROVER_AND",
    Term `!(p:'a->bool) q r. (p AND q) AND r = (p AND r) AND (q AND r)`, 
    pred_TAC) ;

val pOR_ROVER_OR = store_thm 
  ("pOR_ROVER_OR",
    Term `!(p:'a->bool) q r. (p OR q) OR r = (p OR r) OR (q OR r)`, 
    pred_TAC) ;

val pIMP_REFL = store_thm
  ("pIMP_REFL",
    Term `!p:'a->bool. |== (p IMP p)`, 
    pred_TAC) ;

val pIMP_TRANS = store_thm
  ("pIMP_TRANS",
   Term `!(p:'a->bool) q r. |== (p IMP q) /\ |== (q IMP r) ==>
                            |== (p IMP r)`,
   pred_TAC) ;

val pIMP_ANTISYM = store_thm
  ("pIMP_ANTISYM",
    Term `!(p:'a->bool) q. |== (p IMP q) /\ |== (q IMP p) ==>
                           (p=q)`,
    REWRITE_TAC [pSEQ_DEF, pIMP_DEF]   
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN EXT_TAC THEN EQ_TAC THEN PROVE_TAC [] ) ;

val qAND_GLB1 = store_thm
  ("qAND_GLB1",
    Term `!(W:'b->bool) (P:'b->('a->bool)). 
                !i:: W. |== ((!! j:: W. P j) IMP (P i))`,
    pred_TAC) ;

val qAND_GLB2 = store_thm
  ("qAND_GLB2",
    Term `!W (P:'b->('a->bool)) q.
             (!i:: W. |== (q IMP (P i))) ==> |== (q IMP (!! j:: W. P j))`,
    pred_TAC) ;

val qOR_LUB1 = store_thm
  ("qOR_LUB1",
    Term `!W (P:'b->('a->bool)).
             !i:: W. |== ((P i) IMP (?? j:: W. P j))`,
    pred_TAC) ; 


val qOR_LUB2 = store_thm
  ("qOR_LUB2",
    Term `!W (P:'b->('a->bool)) q.
              (!i:: W. |== ((P i) IMP q)) ==> |== ((?? j:: W. P j) IMP q)`,
    pred_TAC) ;

val FF_TOP = store_thm
  ("FF_TOP",
    Term `!p:'a->bool. |== (FF IMP p)`, 
    pred_TAC) ;

val TT_BOTTOM = store_thm
  ("TT_BOTTOM",
    Term `!p:'a->bool. |== (p IMP TT)`, 
    pred_TAC) ;

val qOR_EMPTY = store_thm 
  ("qOR_EMPTY",
    Term `!P:'b->('a->bool). (??i:: FF. P i) = FF`,
    pred_TAC) ;

val qAND_EMPTY = store_thm 
  ("qAND_EMPTY",
    Term `!P:'b->('a->bool). (!!i:: FF. P i) = TT`,
    pred_TAC) ;

val NOT_TT = store_thm
  ("NOT_TT", 
    Term `NOT (TT:'a->bool) = FF`, 
    pred_TAC) ;

val NOT_FF = store_thm
  ("NOT_FF", 
    Term `NOT (FF:'a->bool) = TT`, 
    pred_TAC) ;

val pDEMORGAN1 = store_thm 
  ("pDEMORGAN1", 
    Term `!(p:'a->bool) q. (NOT (p AND q)) = ((NOT p) OR (NOT q))`,
    pred_TAC) ;

val pDEMORGAN2 = store_thm 
  ("pDEMORGAN2", 
    Term `!(p:'a->bool) q. (NOT (p OR q)) = ((NOT p) AND (NOT q))`,
    pred_TAC) ;

val pCONTRA_POS = store_thm
  ("pCONTRA_POS",
    Term `!(p:'a->bool) q. (p IMP q) = ((NOT q) IMP (NOT p))`,
    pred_TAC) ;

val qAND_SINGvalON = store_thm
  ("qAND_SINGvalON",
    Term `!(P:'b->('a->bool)) a. (!!i:: (\j. a=j). P i) = (P a)`,
    pred_TAC) ;

val qOR_SINGvalON = store_thm
  ("qOR_SINGvalON",
    Term `!(P:'b->('a->bool)) a. (??i:: (\j. a=j). P i) = (P a)`,
    pred_TAC) ;

val qAND_AND = store_thm
  ("qAND_AND",
    Term `!(p:'a->bool) q. (!!i:: (\j. (p=j) \/ (q=j)). i) = (p AND q)`,
    pred_TAC) ;

val qOR_OR = store_thm
  ("qOR_OR",
    Term `!(p:'a->bool) q. (??i:: (\j. (p=j) \/ (q=j)). i) = (p OR q)`,
    pred_TAC) ;

val qAND_RIGHT_DISTR_AND = store_thm
  ("qAND_RIGHT_DISTR_AND",
    Term `!W (P:'b->('a->bool)) p. 
               ~(W=FF)  ==>  ((!!i::W. P i) AND p = (!!i::W. (P i) AND p))`,
    pred_TAC) ;

val qAND_LEFT_DISTR_AND = store_thm
  ("qAND_LEFT_DISTR_AND",
    Term `!W (P:'b->('a->bool)) p. 
             ~(W=FF)  ==>  (p AND (!!i::W. P i) = (!!i::W. p AND (P i)))`,
    pred_TAC) ;    

val qAND_LEFT_DISTR_OR = store_thm 
  ("qAND_LEFT_DISTR_OR",
    Term `!W (P:'b->('a->bool)) p. 
              p AND (??i::W. P i) = (??i::W. p AND P i)`,
    pred_TAC ) ;    

val qAND_RIGHT_DISTR_OR = store_thm 
  ("qAND_RIGHT_DISTR_OR",
    Term `!W (P:'b->('a->bool)) p. 
             (??i::W. P i) AND p = (??i::W. (P i) AND p)`,
    pred_TAC ) ;

val qOR_RIGHT_DISTR_AND = store_thm
  ("qOR_RIGHT_DISTR_AND",
    Term `!W (P:'b->('a->bool)) p. 
             (!!i::W. P i) OR p = (!!i::W. (P i) OR p)`,
    pred_TAC ) ;

val qOR_LEFT_DISTR_AND = store_thm
  ("qOR_LEFT_DISTR_AND",
    Term `!W (P:'b->('a->bool)) p. 
              p OR (!!i::W. P i) = (!!i::W. p OR (P i))`,
    pred_TAC ) ; 
   
val qOR_RIGHT_DISTR_OR = store_thm 
  ("qOR_RIGHT_DISTR_OR",
    Term  `!W (P:'b->('a->bool)) p. 
               ~(W=FF) ==> (((??i::W. P i) OR p) = (??i::W. (P i) OR p))`,
    pred_TAC ) ;

val qOR_LEFT_DISTR_OR = store_thm 
  ("qOR_LEFT_DISTR_OR",
    Term `!W (P:'b->('a->bool)) p. 
             ~(W=FF) ==> ((p OR (??i::W. P i)) = (??i::W. p OR (P i)))`,     
    pred_TAC ) ;

val qAND_DOM_SPLIT = store_thm 
  ("qAND_DOM_SPLIT",
    Term `!W (P:'b->('a->bool)) a. 
             ((!!i:: W. P i) AND (P a)) = (!!i::(\j. (W j) \/ (a=j)). P i)`,
    pred_TAC ) ;

val qOR_DOM_SPLIT = store_thm 
  ("qOR_DOM_SPLIT",
    Term `!W (P:'b->('a->bool)) a. 
             ((??i:: W. P i) OR (P a)) = (??i::(\j. (W j) \/ (a=j)). P i)`,
    pred_TAC ) ;

val CHF_FINITE = Define `CHF_FINITE (W:'a->bool) = FINITE (SPEC W)`  ;

val EMPTY_lemma = store_thm
  ("EMPTY_lemma",
   Term `{} = SPEC(FF:'a->bool)`,
   REWRITE_TAC [EMPTY_DEF , FF_DEF]) ;

val CHF_SPEC_lemma = (REWRITE_RULE [] o BETA_RULE o CONJUNCT2) set_ISO_DEF ;

val _ = save_thm ("CHF_SPEC_lemma" , CHF_SPEC_lemma) ;

val SET_lemma = store_thm
  ("SET_lemma",
    Term `!(a:'a) V. {y| (y=a) \/ (y IN V)} = SPEC(\y.(y=a) \/ ((CHF V) y))`,
   REWRITE_TAC [IN_DEF, GSPEC_DEF]
   THEN BETA_TAC THEN REWRITE_TAC [pairTheory.PAIR_EQ]
   THEN REPEAT GEN_TAC
   THEN AP_TERM_TAC 
   THEN pred_TAC ) ;


val lemma1 = prove
   (Term  `(!V a. X V ==> X(\x:'a. (x=a) \/ (V x))) ==>
           (!s:('a)set. X(CHF s) ==> (!e:'a. X(\y. (y=e) \/ (CHF s y))))`,
    RW_TAC bool_ss []) ;    

val lemma2 = prove
   (Term `(!s':('a)set. 
               P s' ==> 
               (!e:'a. P(SPEC(\y. (y=e) \/ (CHF s' y)))))  ==>
               (!V (a:'a). P(SPEC V) ==> P(SPEC(\x:'a. (x=a) \/ V x)))`,
    REPEAT STRIP_TAC
    THEN RES_TAC 
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [CHF_SPEC_lemma])
    THEN ASM_REWRITE_TAC[]) ;

val CHF_FINITE_DEF = store_thm 
  ("CHF_FINITE_DEF",
   Term `!W:'a->bool. (CHF_FINITE W) =
              (!X. (X FF) /\ (!V a. (X V) ==> 
              (X (\x. (x=a) \/ (V x))))  ==> (X W))`,
   REWRITE_TAC [CHF_FINITE,FINITE_DEF,INSERT_DEF,
                SET_lemma,EMPTY_lemma] 
   THEN GEN_TAC
   THEN EQ_TAC THEN REPEAT STRIP_TAC
   THENL 
   [ IMP_RES_TAC lemma1
     THEN XRULE_ASSUM_TAC  
          ((REWRITE_RULE [CHF_SPEC_lemma]) o  BETA_RULE o 
           (SPEC (Term `\x:('a)set. (X:('a->bool)->bool) (CHF x)`)))
     THEN RES_TAC ,
     XRULE_ASSUM_TAC 
          (BETA_RULE o (SPEC (Term `\x:'a->bool. (P:('a)set->bool)(SPEC x)`)))
     THEN FIRST_ASSUM MATCH_MP_TAC
     THEN ASM_REWRITE_TAC []
     THEN REPEAT STRIP_TAC
     THEN RES_TAC 
     THEN RULE_ASSUM_TAC (REWRITE_RULE [CHF_SPEC_lemma])
     THEN ASM_REWRITE_TAC [] 
    ]) ;


val CHF_INDUCTION = store_thm 
  ("CHF_INDUCTION",
    Term `!(X:('b->bool)->bool). 
                 ((X FF) /\ (!V a. (X V) ==> X(\x. (x=a) \/ (V x))))
                 ==>
                 (!V. (CHF_FINITE  V) ==> X V)`,
   REWRITE_TAC [CHF_FINITE_DEF] 
   THEN REPEAT STRIP_TAC
   THEN RES_TAC ) ;

val _ = export_theory() ; 
