(*-------------------------------------------------------------------
  A theory about relations.

  A relation between type A and B is represented by type A->B->bool.

  Notions defined:

  - LinRel: linear relation
  - SUBREL: sub relation
  - rUNION: union of two relations
  - rINTER: intersection of two relations
  - rSEQ : composition between two relations
  - rRAISE : used to define R o R o R ... o R, n times
  - WOK : reversing the direction of a relation
  - rId : the identity relation 
  - rBOT : the empty relation
  - rTOP : the full relation
  - TRANS : transitive relation
  - RELF : reflexive relation
  - ReflC: reflexive closure
  - ANTI_SYM : anti-symmetric relation
  - ANTI_REFL : anti-reflexive relation

 theorems:
 
  - some theorems about reflexive closures
  - extensionality of relations

  Author: Wishnu Prasetya
  Last upadate: march 1999

-----------------------------------------------------------------*)

open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;
  
val _ = new_theory "urel" ;


(* load "bossLib" ; *)

open bossLib ;


val LinRel = Define
   `LinRel r = !(a:'a) b. r a b \/ (a=b) \/ r b a` ;

val SUBREL = Define
   `SUBREL r s = (!(x:'a) (y:'b). r x y ==> s x y)` ;
val _ = set_fixity "SUBREL" (Infix 540) ;

val rUNION = Define
   `rUNION r s  =  \(x:'a) (y:'b). r x y \/ s x y` ;
val _ = set_fixity "rUNION" (Infix 580) ;

val rINTER = Define
   `rINTER r s  =  \(x:'a) (y:'b). r x y /\ s x y` ;
val _ = set_fixity "rINTER" (Infix 600) ;

val rSEQ = Define
   `rSEQ (R1:'a->'b->bool) (R2:'b->'c->bool) x z = (?y. R1 x y /\ R2 y z)` ;
val _ = set_fixity "rSEQ" (Infix 560) ;

val WOK = Define
   `WOK (R:'a->'b->bool)  =  \x y. R x y` ;

val rId = Define
   `(rId:'a->'a->bool) =  \x y. (x=y)` ;

val rBOT = Define
   `(rBOT:'a->'b->bool) = \x y. F` ;

val rTOP = Define
   `(rTOP:'a->'b->bool) = \x y. T` ;

val TRANS_DEF = Define
   `TRANS r = (!(x:'a) y z. r x y /\ r y z ==> r x z)` ;

val REFL_DEF = Define
   `REFL r = (!(x:'a). r x x)` ;

val ANTI_REFL = Define
   `ANTI_REFL r = (!(x:'a). ~r x x)` ;

val ANTI_SYM = Define
   `ANTI_SYM r = (!(x:'a) y. r x y /\ r y x ==> (x=y))` ;

(*---------------------------------------------------------------------------
                               n              
               definition of  R
---------------------------------------------------------------------------*)

val rRAISE = Define
   `(rRAISE (R:'a->'a->bool) 0 = (\x y. (x=y))) /\
    (rRAISE  R (SUC n)         = (R rSEQ (rRAISE R n)))` ;

(*---------------------------------------------------------------------------
               reflexive closure
---------------------------------------------------------------------------*)

val ReflC = Define
   `ReflC (R:'a->'a->bool) x y = R x y \/ (x=y)` ;

(*---------------------------------------------------------------------------
    Theorems part
---------------------------------------------------------------------------*)

val WRel = ty_antiq (Type `:'a->'a->bool`) ;

fun DEL_ALL_TAC (asml,gl) =
    if (asml=[]) then ALL_TAC (asml,gl)
    else (POP_ASSUM (fn thm=> ALL_TAC) THEN DEL_ALL_TAC) (asml,gl) ;

val ANTI_SYM_ReflC = store_thm
  ("ANTI_SYM_ReflC",
   Term `TRANS (r:^WRel) /\ ANTI_REFL r ==> ANTI_SYM (ReflC r)`,
    REWRITE_TAC [TRANS_DEF, ANTI_REFL, ANTI_SYM, ReflC]
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN RES_TAC THEN RES_TAC) ;

val TRANS_ReflC = store_thm
  ("TRANS_ReflC",
   Term `TRANS (r:'a->'a->bool) ==> TRANS (ReflC r)`,
    REWRITE_TAC [TRANS_DEF, ReflC]
    THEN DEL_ALL_TAC THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THENL [ RES_TAC THEN ASM_REWRITE_TAC[] ,
           FIRST_ASSUM (SUBST1_TAC o SYM) THEN ASM_REWRITE_TAC[] ]) ;

val ReflC_R_ReflC_TRANS = store_thm
  ("ReflC_R_ReflC_TRANS",
   Term `!R (a:'a) b c. 
       TRANS R /\ ReflC R a b /\ R b c ==> ReflC R a c`,
    REWRITE_TAC [TRANS_DEF, ReflC] THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[]) ;

val R_ReflC_ReflC_TRANS = store_thm
  ("R_ReflC_ReflC_TRANS",
   Term `!R (a:'a) b c. 
       TRANS R /\ R a b /\ ReflC R b c ==> ReflC R a c`,
    REWRITE_TAC [TRANS_DEF, ReflC] THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[]
    THEN FIRST_ASSUM (fn thm=> REWRITE_TAC [SYM thm])
    THEN ASM_REWRITE_TAC[]) ;

val ReflC_R_R_TRANS = store_thm
  ("ReflC_R_R_TRANS",
   Term `!R (a:'a) b c. 
       TRANS R /\ ReflC R a b /\ R b c ==> R a c`,
    REWRITE_TAC [TRANS_DEF, ReflC] THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[] 
    THEN FIRST_ASSUM (fn thm=> REWRITE_TAC [SYM thm])
    THEN ASM_REWRITE_TAC[]) ;

val R_ReflC_R_TRANS = store_thm
  ("R_ReflC_R_TRANS",
   Term `!R (a:'a) b c. 
       TRANS R /\ R a b /\ ReflC R b c ==> R a c`,
    REWRITE_TAC [TRANS_DEF, ReflC] THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN ASM_REWRITE_TAC[] 
    THEN FIRST_ASSUM (fn thm=> REWRITE_TAC [SYM thm])
    THEN ASM_REWRITE_TAC[]) ;

(*-----------------------------------------------------------------------
  Here are extension theorems for relations. Can be used to store 
  equality of two elements in the domain.
-----------------------------------------------------------------------*)

fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;    

val Low_EXTENSION = store_thm
  ("Low_EXTENSION",
   Term `!r x (y:'a). REFL r /\ ANTI_SYM r ==> 
                ((x=y) = (!z. r z x = r z y))`,
    REWRITE_TAC [REFL_DEF, ANTI_SYM]
    THEN REPEAT STRIP_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL [ ASM_REWRITE_TAC[] , ALL_TAC ]
    THEN FIRST_ASSUM MATCH_MP_TAC THEN CONJ_TAC
    THENL
    [ XRULE_ASSUM_TAC (SPEC (Term `x:'a`)) THEN RES_TAC ,
      XRULE_ASSUM_TAC (SPEC (Term `y:'a`)) THEN RES_TAC ]) ;

val High_EXTENSION = store_thm
  ("High_EXTENSION",
   Term `!r x (y:'a). REFL r /\ ANTI_SYM r ==> 
                ((x=y) = (!z. r x z = r y z))`,
    REWRITE_TAC [REFL_DEF, ANTI_SYM]
    THEN REPEAT STRIP_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL [ ASM_REWRITE_TAC[] , ALL_TAC ]
    THEN FIRST_ASSUM MATCH_MP_TAC THEN CONJ_TAC
    THENL
    [ XRULE_ASSUM_TAC (SPEC (Term `y:'a`)) THEN RES_TAC ,
      XRULE_ASSUM_TAC (SPEC (Term `x:'a`)) THEN RES_TAC ]) ;


val _ = export_theory() ; 






