(*-----------------------------------------------------------------------------------

  This library contains some basic theorems about well-founded relations. 

  Well foundedness is defined as follows. A relation R over type `a is called 
  well founded iff it satisfies:

        !V. (!y. y IN V ==> (?x. R x y /\ x IN V)) ==> (V = {})

  That is, if one can indefinitely decrease in R inside a set V, then V is empty. 
  In other words, a never ending decrease is not possible.

  We first give a more general definition of well-foundedness of R over a subset
  A of type `a:

        !V. (!y. y IN (V INTER A) ==> (?x. R x y /\ x IN (V INTER A)) 
             ==> 
            (V INTER A = {})

  The first definition can be obtained by instantiating A to UNIV.

  Definitions and theorems related with this more general definition is usually
  marked with 's'. For example above definition is called 'sWF' whereas the first
  above is called 'WELL_FOUNDED'

  The main results of this library is as follows (there is also an 'sWF' version
  of almost every result mentioned below):

     - WELL_FOUNDED
       The definition for well-foundedness.
     
     - ADMIT_WF_INDUCTION
       Define the well-founded induction.

     - WF_EQU_IND
       States that well-foundedness is equivalent with admiting the WF induction.
 
     - FIN_ADMIT_WF_IND        
       States that a transitive and anti-reflexive relation on finite domain will
       admit WF induction.

     - SUB_WF
       If A is a subset of B and R restricted on B is well founded, then so is
       R restricted on A.

     - NUM_ADMIT_WF_INDUCTION
       States that the relation < on natural number admits WF induction.

     - REL_JOINT_ADMIT_WF_IND
       States that the conjunction of a WF relation with another relation is WF.

     - REL_PROD_ADMIT_WF_IND
       States that the product a of WF relation with a nother relation is WF.

     - REL_LEXI_ADMIT_WF_IND
       States that the lexicographic product of two WF relations is WF.

     - FPROD_ADMIT_WF_IND
       States that the finite product of relations is WF if at least one of 
       the original relation is WF.

      - FLEXI_ADMIT_WF_IND
        States that the finite lexicographic product of WF relations is again
        WF.

  Lib name: uwf
  Requires libraries: upred and urel.
  Last Update : april 99
  Author  : Wishnu Prasetya
            Dept. of Comp. Science
            University of Utrecht, the Netherlands
  EMail   : wishnu@cs.uu.nl

--------------------------------------------------------------------------------*)

open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ; 

  
val _ = new_theory "uwf" ;

(*
load "bossLib" ; 
load "upredTheory" ;
load "urelTheory" ;
load "setLib" ;
*)

open bossLib ;
open upredTheory ;
open urelTheory ;
open setTheory ;
open setLib ;
open arithmeticTheory ;

val Rel  = ty_antiq (Type `:'a->'a->bool`) ;
val Rel1 = ty_antiq (Type `:'a->'a->bool`) ;
val Rel2 = ty_antiq (Type `:'b->'b->bool`) ;
val Pred = ty_antiq (Type `:'a->bool`) ;

fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;

local fun EXT_SPEC_TAC (asml,g) = REWRITE_TAC [FUN_EQ_CONV g] (asml,g)
in
val EXT_TAC  =
    REPEAT GEN_TAC 
    THEN EXT_SPEC_TAC 
    THEN BETA_TAC 
    THEN GEN_TAC 
end ;

fun REPEAT_FIN n tac = if n<=0 then ALL_TAC 
                               else (tac THEN REPEAT_FIN (n-1) tac) ;

fun UNDISCH_ALL_TAC (nil,g) = ALL_TAC (nil,g)
  | UNDISCH_ALL_TAC(a::s,g) = (UNDISCH_TAC a THEN UNDISCH_ALL_TAC) (a::s,g) ;

fun DEL_TAC trm = 
    UNDISCH_TAC trm
    THEN DISCH_TAC
    THEN POP_ASSUM (fn thm=> ALL_TAC) ; 


(*---------------------------------------------------------------------------------
               definition of well founded induction
---------------------------------------------------------------------------------*)

val ADMIT_sWF_Ind = Define
   `ADMIT_sWF_Ind V (R:^Rel) =
    !P. (!y. y IN V /\ (!x. x IN V /\ R x y ==> P x) ==> P y) 
        ==> 
        (!x. x IN V ==> P x)` ;

val ADMIT_WF_INDUCTION = Define
   `ADMIT_WF_INDUCTION (R:^Rel) =
    !P. (!y. (!x. R x y ==> P x) ==> P y) ==> (!x. P x)` ;

(*---------------------------------------------------------------------------------
        ADMIT_WF_INDUCTION can be expressed in terms of ADMIT_sWF_Ind
---------------------------------------------------------------------------------*)

val ADMIT_sWF_WF = store_thm
  ("ADMIT_sWF_WF",
   Term `!(R:^Rel). ADMIT_WF_INDUCTION R = ADMIT_sWF_Ind UNIV R`,
    REWRITE_TAC [ADMIT_WF_INDUCTION, ADMIT_sWF_Ind, IN_UNIV]) ;

(*---------------------------------------------------------------------------------
        Definition of well-foundedness (in terms of predicates)
---------------------------------------------------------------------------------*)

val CHF_sWF = Define
   `sWF V (R:^Rel) = 
    (!P. (!y. y IN V /\ P y ==> (?x. x IN V /\ R x y /\ P x)) 
         ==> 
         ~(?x. x IN V /\ P x))` ;

val CHF_WELL_FOUNDED = Define
   `WELL_FOUNDED (R:^Rel) = 
    (!P. (!y. P y ==> (?x. R x y /\ P x)) ==> (P=FF))` ;

val FF_not = prove
  (Term `(p = FF) = ~(?x:'a. p x)`,
    REWRITE_TAC [FF_def] 
    THEN (CONV_TAC o DEPTH_CONV) FUN_EQ_CONV THEN BETA_TAC
    THEN (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
    THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL [ RES_TAC ,ASM_REWRITE_TAC[]]) ;

(*---------------------------------------------------------------------------------
                WELL_FOUNDED can be expressed in terms of sWF
---------------------------------------------------------------------------------*)

val CHF_sWF_WF = store_thm
  ("CHF_sWF_WF",
   Term `!(R:^Rel). WELL_FOUNDED R = sWF UNIV R`,
    REWRITE_TAC [CHF_WELL_FOUNDED, FF_not, CHF_sWF, IN_UNIV]) ;

(*---------------------------------------------------------------------------------
  CONTRA_lemma: |- !p. p /\ ~p ==> F
  The lemma is useful in proofs using contradiction
---------------------------------------------------------------------------------*)

val CONTRA_lemma = store_thm
  ("CONTRA_lemma",Term `!p. (p /\ ~p) ==> F`, PROVE_TAC[]) ;

(*---------------------------------------------------------------------------------
   (Alternative, and actually the standard) definition of well-foundedness in 
   terms of sets.
---------------------------------------------------------------------------------*)

val EXT_lemma = prove
  (Term `!(f:'a->'b) g. (f=g) = (!x. f x = g x)`,
    GEN_TAC THEN
    REWRITE_TAC [FUN_EQ_CONV (Term `(f:'a->'b)=g`)]) ;

val EMPTY_not = prove
  (Term `(V = {}) = ~(?x:'a. x IN V)`,
    REWRITE_TAC [EXTENSION, NOT_IN_EMPTY]
    THEN (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
    THEN REFL_TAC) ;

val sWF = store_thm
  ("sWF",
   Term `!V (R:^Rel). sWF V R =
       (!A. (!y. y IN V /\ y IN A ==> (?x. x IN V /\ R x y /\ x IN A)) 
            ==> 
           (V INTER A ={}))`,
    REWRITE_TAC [CHF_sWF, EMPTY_not, IN_INTER] 
    THEN REPEAT GEN_TAC THEN EQ_TAC
    THEN REPEAT STRIP_TAC
    THENL [ XRULE_ASSUM_TAC (BETA_RULE o SPEC (Term `\s:'a. s IN A`))
            THEN RES_TAC
            THEN XRULE_ASSUM_TAC (BETA_RULE o REWRITE_RULE [FF_def, EXT_lemma])
            THEN ASM_REWRITE_TAC [EXTENSION , NOT_IN_EMPTY] ,
            XRULE_ASSUM_TAC (SPEC (Term `{s:'a | P s}`))
            THEN (XRULE_ASSUM_TAC o CONV_RULE o DEPTH_CONV) SET_SPEC_CONV
            THEN RES_TAC ] ) ;
   
val WELL_FOUNDED = store_thm
  ("WELL_FOUNDED",
   Term `!R:^Rel. WELL_FOUNDED R =
       (!V. (!y. y IN V ==> (?x. R x y /\ x IN V)) ==> (V={}))`,
    REWRITE_TAC [ CHF_sWF_WF, sWF, IN_UNIV, INTER_UNIV]) ;   

(*---------------------------------------------------------------------------------
               The meaning of anti-well-foundedness.
---------------------------------------------------------------------------------*)

val NOT_sWF = store_thm
  ("NOT_sWF",
   Term `!V (R:^Rel). ~(sWF V R) =
      (?P. (!y. y IN V /\ P y ==> (?x. x IN V /\ R x y /\ P x))
           /\ (?x. x IN V /\ P x))`,
    REWRITE_TAC [CHF_sWF]
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REWRITE_TAC [NOT_IMP]) ;

val NOT_WF = store_thm
  ("NOT_WF",
   Term `!R:^Rel. ~(WELL_FOUNDED R) =
      (?P. (!y. P y ==> (?x. R x y /\ P x)) /\ ~(P=FF))`,
    REWRITE_TAC [ CHF_sWF_WF, NOT_sWF, FF_not, IN_UNIV]) ;

(*---------------------------------------------------------------------------------
  WELL_FOUNDED_CONTRA: The contra position version of WELL_FOUNDED definition.
---------------------------------------------------------------------------------*)

val sWF_CONTRA = store_thm
  ("sWF_CONTRA",
   Term `!V (R:^Rel). sWF V R =
      (!P. (?x. x IN V /\ P x) 
            ==> 
            ?y. y IN V /\ P y /\ (!x. x IN V /\ P x ==> ~(R x y)))`,
   REWRITE_TAC [CHF_sWF]
   THEN REPEAT GEN_TAC  
   THEN (CONV_TAC o RATOR_CONV o RAND_CONV o RAND_CONV o ABS_CONV) CONTRAPOS_CONV
   THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
   THEN REWRITE_TAC [NOT_IMP]
   THEN (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
   THEN REWRITE_TAC [PROVE [] `~(p /\ q /\ r) = p /\ r ==> ~q`]
   THEN REWRITE_TAC [PROVE [] `(p /\ q) /\ r  = p /\ ( q /\ r)`]) ;

val WELL_FOUNDED_CONTRA = store_thm
  ("WELL_FOUNDED_CONTRA",
   Term `!R:^Rel. WELL_FOUNDED R =
      (!P. ~(P=FF) ==> ?y. P y /\ (!x. P x ==> ~(R x y)))`,
    REWRITE_TAC [ CHF_sWF_WF, sWF_CONTRA, FF_not, IN_UNIV]) ;

(*--------------------------------------------------------------------------------
  FF_lemma: |- !P. (P = FF) = (!x. ~P x)
--------------------------------------------------------------------------------*)

val FF_lemma = store_thm
  ("FF_lemma", 
   Term `!P:^Pred. (P=FF) = (!x. ~P x)`, 
    REWRITE_TAC [FF_def]
    THEN GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
    THENL [RULE_ASSUM_TAC 
                (fn thm=> liteLib.orelsef
                          (BETA_RULE o SUBS [ASSUME (Term `(P:^Pred)=(\s. F)`)], 
                           I) 
                           thm) 
           THEN ASM_REWRITE_TAC[] ,
           EXT_TAC THEN ASM_REWRITE_TAC[] ]) ;

(*--------------------------------------------------------------------------------
  Now we are going to prove that well foundedness is equal with admiting WF
  induction.
--------------------------------------------------------------------------------*)

val lemma1 = PROVE []
   `!V (R:^Rel) P.
    (!y. y IN V /\ (!x. x IN V /\ R x y ==> ~P x) ==> ~P y) = 
    (!y. y IN V /\ P y ==> (?x. x IN V /\ R x y /\ P x))` ;


val lemma2 = PROVE []
   `!V (R:^Rel) P.
    (!y. y IN V /\ ~P y ==> (?x. x IN V /\ R x y /\ ~P x)) = 
    (!y. y IN V /\ (!x. x IN V /\ R x y ==> P x) ==> P y)` ;

val sWF_EQU_IND = store_thm
  ("sWF_EQU_IND",
   Term `!V (R:^Rel). (sWF V R) = (ADMIT_sWF_Ind V R)`,
    REWRITE_TAC [CHF_sWF, ADMIT_sWF_Ind, FF_lemma]
    THEN REPEAT GEN_TAC THEN EQ_TAC
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC 
         (BETA_RULE o REWRITE_RULE [NOT_def] o SPEC (Term `NOT (P:^Pred)`))
    THENL [ XRULE_ASSUM_TAC (REWRITE_RULE [lemma2])
            THEN RES_TAC THEN UNDISCH_TAC (Term `~P (x:'a) ==> F`)
            THEN PROVE_TAC[] ,
            RULE_ASSUM_TAC (REWRITE_RULE [lemma1])
            THEN RES_TAC ]) ;

val WF_EQU_IND = store_thm
  ("WF_EQU_IND",
   Term `!(R:^Rel). WELL_FOUNDED R = ADMIT_WF_INDUCTION R`,
    REWRITE_TAC [CHF_sWF_WF, ADMIT_sWF_WF, sWF_EQU_IND]) ;

(*-------------------------------------------------------------------------------
  now we will prove that a transitive and anti-reflexive relation on a finite
  domain will admit WF induction.

  This is proved using induction, as finiteness implies induction. Val ALL_WF be
  defined as follows:

      ALL_WF V R A =   !B. B SUBSET A ==> WF V R B
      WF V R A     =  (!y:'a. y IN (V INTER A) ==> ?x. R x y /\ x IN (V INTER A)) 
                       ==> 
                      (V INTER A = {})

  The proof is based on three lemmas:

      FIN_WF_lemma1: States that ALL_WF UNV implies well foundedness, hence 
                     also the WF induction.
      
      FIN_WF_lemma2: States that ALL_WF R holds for empty set.
      
      FIN_WF_lemma3: States that if ALL_WF R holds for set V, then it also
                     holds for a INSERT V.

-------------------------------------------------------------------------------*)

val WF = Term `\V R A. (!y:'a. y IN V /\ y IN A ==> ?x. x IN V /\ R x y /\ x IN A) 
                   ==> (V INTER A ={})` ;

val ALL_WF = Term `\V R A. !B. B SUBSET A ==> ^WF V R B` ;

val FIN_WF_lemma1 = (BETA_RULE o prove)
  (Term `!V (R:^Rel). ^ALL_WF V R V ==> sWF V R`,
    REWRITE_TAC [sWF] THEN BETA_TAC
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC (Term `(V:('a)set) INTER A`))
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [INTER_SUBSET, IN_INTER])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [SYM (SPEC_ALL INTER_ASSOC)])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [INTER_IDEMPOT])
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC 
    THEN RES_TAC THEN EXISTS_TAC (Term `x':'a`)
    THEN ASM_REWRITE_TAC[]) ;

val FIN_WF_lemma2 = (BETA_RULE o prove)
  (Term `!V (R:^Rel). ^ALL_WF V R {}`,
   BETA_TAC 
   THEN REWRITE_TAC [SUBSET_EMPTY]
   THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC [INTER_EMPTY]) ;

val FIN_WF_lemma3_1 = prove
  (Term `!V (R:^Rel) A. 
       ((!y. y IN V /\ y IN A ==> (?x'. x' IN V /\ R x' y /\ x' IN A)) 
        ==> (V INTER A = {})) 
        =
       (~(V INTER A = {}) ==> 
       (?y. y IN V /\ y IN A /\ (!x'. x' IN V /\ x' IN A ==> ~R x' y)))`,
    REPEAT GEN_TAC
    THEN (CONV_TAC o RAND_CONV o ONCE_DEPTH_CONV) CONTRAPOS_CONV
    THEN (CONV_TAC o DEPTH_CONV) NOT_EXISTS_CONV
    THEN REWRITE_TAC [DECIDE `~(p /\ q /\ r) = (p /\ q ==> ~r)`]
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REWRITE_TAC [DECIDE `~(p /\ q ==> ~r) = (p /\ r /\ q)`]) ;

val FIN_WF_lemma3_2 = prove
  (Term `!V (R:^Rel) A x.
       (?y. y IN V /\ (y = x) /\ (!x'. x' IN V /\ ((x' = x) \/ x' IN (A DEVALE x)) 
                                        ==> ~R x' y))
         =
       (x IN V /\ (!x'. x' IN V /\ x' IN (A DEVALE x) ==> ~R x' x) /\ ~R x x)`,
    REPEAT STRIP_TAC THEN EQ_TAC
    THEN REPEAT STRIP_TAC
    THENL [ FIRST_ASSUM (SUBST1_TAC o SYM) THEN ASM_REWRITE_TAC[] ,
            XRULE_ASSUM_TAC (SUBS [SYM (ASSUME (Term `y=(x:'a)`))])
            THEN RES_TAC ,
            XRULE_ASSUM_TAC (SUBS [SYM (ASSUME (Term `y=(x:'a)`))])
            THEN RES_TAC ,
            EXISTS_TAC (Term `x:'a`) THEN REWRITE_TAC[]
            THEN REPEAT STRIP_TAC
            THENL [ ASM_REWRITE_TAC[] ,
                    XRULE_ASSUM_TAC (SUBS [SYM (ASSUME (Term `x'=(x:'a)`))])
                    THEN RES_TAC ,
                    RES_TAC ]
           ]) ;

val FIN_WF_lemma3_3 = prove
  (Term `!V (R:^Rel) A x. 
          ~(!x'. x' IN V /\ x' IN (A DEVALE x) ==> ~R x' x) = 
           (?x'. x' IN V /\ x' IN (A DEVALE x) /\  R x' x)`,
    (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN REWRITE_TAC 
         [DECIDE `!p q. ~(p /\ q==> ~r) = (p /\ q /\ r)`] );

val FIN_WF_lemma3 = (BETA_RULE o prove)
  (Term `!V (R:^Rel) A x.
       TRANS R /\ ANTI_REFL R /\ ^ALL_WF V R A 
       ==> 
       ^ALL_WF V R (x INSERT A)`,
    BETA_TAC THEN REWRITE_TAC [TRANS_def, ANTI_REFL_def, FIN_WF_lemma3_1]
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [SUBSET_INSERT_DELETE])
    THEN ASM_CASES_TAC (Term `(x:'a) IN B`)
    THENL 
      [ (*- 1 -*)
       (SUBST1_TAC o SYM  o UNDISCH) (SPECL [Term `x:'a`, Term `B:('a)set`] 
                                      INSERT_DELETE)
       THEN REWRITE_TAC [IN_INSERT]
       THEN REWRITE_TAC 
            [DECIDE `p/\(q1\/q2)/\r=(p/\q1/\r)\/(p/\q2/\r)`]
       THEN CONV_TAC EXISTS_OR_CONV
       THEN REWRITE_TAC [FIN_WF_lemma3_2]
       THEN ASM_REWRITE_TAC []
       THEN REWRITE_TAC [DECIDE `p/\q\/r=~p\/~q==>r`]
       THEN REWRITE_TAC [FIN_WF_lemma3_3]
       THEN REPEAT STRIP_TAC 
       THENL
       [ (*- 1.1 -*)
         SUBGOAL_THEN (Term `~(V INTER (B DELETE (x:'a)) = {})`) ASSUME_TAC
         THENL 
         [ UNDISCH_TAC (Term `~(V INTER (B:('a)set) = {})`)
           THEN REWRITE_TAC [SYM (SPEC_ALL MEMBER_NOT_EMPTY)]
           THEN REWRITE_TAC [IN_INTER, IN_DELETE]
           THEN REPEAT STRIP_TAC THEN EXISTS_TAC (Term `x':'a`)
           THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
           THEN XRULE_ASSUM_TAC (SUBS [ASSUME (Term `x'=(x:'a)`)]) THEN RES_TAC ,
           ALL_TAC ]
         THEN RES_TAC
         THEN EXISTS_TAC (Term `y:'a`) THEN ASM_REWRITE_TAC []
         THEN REPEAT STRIP_TAC
         THENL
         [ XRULE_ASSUM_TAC (SUBS [ASSUME (Term `x'=(x:'a)`)]) THEN RES_TAC ,
           RES_TAC ] ,
         (*- 1.2 -*)
         SUBGOAL_THEN (Term `~(V INTER (B DELETE (x:'a)) = {})`) ASSUME_TAC
         THENL 
         [ REWRITE_TAC [SYM (SPEC_ALL MEMBER_NOT_EMPTY)]
           THEN EXISTS_TAC (Term `x':'a`) THEN ASM_REWRITE_TAC [IN_INTER] ,
           ALL_TAC ]
         THEN RES_TAC
         THEN EXISTS_TAC (Term `y:'a`) THEN ASM_REWRITE_TAC []
         THEN REPEAT STRIP_TAC
         THENL
         [ XRULE_ASSUM_TAC (SUBS [ASSUME (Term `x''=(x:'a)`)]) THEN RES_TAC ,
           RES_TAC ] ] ,
      (*- 2 -*)
      IMP_RES_TAC DELETE_NON_ELEMENT
      THEN SUBGOAL_THEN (Term `~(V INTER (B DELETE (x:'a)) = {})`) ASSUME_TAC
      THENL [ ASM_REWRITE_TAC [] , ALL_TAC ]
      THEN FIRST_ASSUM (SUBST1_TAC o SYM)
      THEN RES_TAC 
      THEN EXISTS_TAC (Term `y:'a`) THEN ASM_REWRITE_TAC []
    ]) ;

(*-------------------------------------------------------------------------------
  FIN_ADMIT_WF_IND: 
  States that a transitive and anti reflexive relation on a finite domain will
  admit WF induction.
-------------------------------------------------------------------------------*)

val FIN_ADMIT_sWF_Ind = store_thm
  ("FIN_ADMIT_sWF_Ind",
   Term `!V (R:^Rel).
        TRANS R  /\  ANTI_REFL R  /\  FINITE V
        ==>
        ADMIT_sWF_Ind V R`,
    REWRITE_TAC [(SYM o SPEC_ALL) sWF_EQU_IND]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC FIN_WF_lemma1
    THEN MATCH_MP_TAC 
         ((BETA_RULE o SPEC (Term `^ALL_WF V (R:^Rel)`) o 
           REWRITE_RULE [FINITE_DEF] o 
           ASSUME) (Term `FINITE (V:('a)set)`))
    THEN REPEAT STRIP_TAC
    THENL [ IMP_RES_TAC FIN_WF_lemma2 ,
            IMP_RES_TAC FIN_WF_lemma3 ] ) ;

val FIN_ADMIT_WF_IND = store_thm
  ("FIN_ADMIT_WF_IND",
   Term `!(R:^Rel).
        TRANS R  /\  ANTI_REFL R  /\  FINITE (UNIV:('a)set)
        ==>
        ADMIT_WF_INDUCTION R`,
    REWRITE_TAC [ ADMIT_sWF_WF, FIN_ADMIT_sWF_Ind]) ;

(*-------------------------------------------------------------------------------
  If (A,R) is well founded, then a sub-relation (B,S) of (A,R) is also 
  well founded.
-------------------------------------------------------------------------------*)

val lemma = prove
  (Term `A2 SUBSET A1 /\ (A1 INTER (A2 INTER A) = {}) ==> 
         ((A2:('a)set) INTER A = {})`,
    REWRITE_TAC [SUBSET_DEF, EXTENSION, NOT_IN_EMPTY, IN_INTER]
    THEN REPEAT STRIP_TAC
    THEN REPEAT RES_TAC) ;

val SUB_WF = store_thm
  ("SUB_WF",
   Term `!A1 A2 (R1:^Rel) (R2:^Rel).
       sWF A1 R1 /\ A2 SUBSET A1 /\ (R2 SUBREL R1) ==> sWF A2 R2`,
    REWRITE_TAC [SUBREL_def, SUBSET_DEF, sWF]
    THEN REPEAT STRIP_TAC
    THEN MATCH_MP_TAC lemma
    THEN ASM_REWRITE_TAC [SUBSET_DEF]
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC [IN_INTER]
    THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN EXISTS_TAC (Term `x:'a`) THEN RES_TAC
    THEN ASM_REWRITE_TAC[]) ;

(*-------------------------------------------------------------------------------
  ADMIT_WF_ANTI_REFL = |- !R. ADMIT_WF_INDUCTION R ==> ANTI_REFL R
-------------------------------------------------------------------------------*)

val ADMIT_sWF_ANTI_REFL = store_thm
  ("ADMIT_sWF_ANTI_REFL",
   Term `!V (R:^Rel). ADMIT_sWF_Ind V R ==> (!x. x IN V ==> ~R x x)`,
    REWRITE_TAC [ ADMIT_sWF_Ind]
    THEN REPEAT GEN_TAC THEN STRIP_TAC
    THEN XRULE_ASSUM_TAC (BETA_RULE o SPEC (Term `\x:'a. ~R x x`))
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC
    THEN RES_TAC) ;

val ADMIT_WF_ANTI_REFL = store_thm
  ("ADMIT_WF_ANTI_REFL",
   Term `!(R:^Rel). ADMIT_WF_INDUCTION R ==> ANTI_REFL R`,
    let val lemma = REWRITE_RULE [IN_UNIV] 
                (SPEC (Term `UNIV:('a)set`) ADMIT_sWF_ANTI_REFL)
    in
    REWRITE_TAC [ ADMIT_sWF_WF, ANTI_REFL_def, lemma] 
    end 
    ) ;

(*-------------------------------------------------------------------------------
   Prove that < on natural number admits WF_INDUCTION. First we need the 
   following lemmas.
-------------------------------------------------------------------------------*)

val LESS_EQ__EQ_LESS_SUC = DECIDE `!x y. (x <= y) = (x < (SUC y))` ;

 
val NUM_ADMIT_lemma1 = prove
  (Term `!P z. (!y. (!x. x<y ==> P x) ==> P y)  /\  (!x. x<=z ==> P x)
          ==> (!x. x <= (SUC z) ==> P x)`,
    REPEAT GEN_TAC THEN STRIP_TAC 
    THEN RULE_ASSUM_TAC (REWRITE_RULE [LESS_EQ__EQ_LESS_SUC])
    THEN REWRITE_TAC [LESS_OR_EQ] 
    THEN REPEAT STRIP_TAC
    THEN RES_TAC THEN ASM_REWRITE_TAC[]) ;

val NUM_ADMIT_lemma2 = prove
  (Term `!P. (!y. (!x. x<y ==> P x) ==> P y) ==> (!x. x <= 0 ==> P x)`,
    REPEAT GEN_TAC THEN STRIP_TAC
    THEN REWRITE_TAC [LESS_OR_EQ]
    THEN REPEAT STRIP_TAC
    THENL [ RULE_ASSUM_TAC (REWRITE_RULE [prim_recTheory.NOT_LESS_0]) 
            THEN CONTR_TAC (ASSUME (Term `F`)) ,
            XRULE_ASSUM_TAC (SPEC (Term `0`))
            THEN RULE_ASSUM_TAC (REWRITE_RULE [prim_recTheory.NOT_LESS_0]) 
            THEN ASM_REWRITE_TAC[] ]) ;

(*-------------------------------------------------------------------------------
  NUM_ADMIT_WF_INDUCTION: states that < on num admits WF_INDUCTION
-------------------------------------------------------------------------------*)

val NUM_ADMIT_WF_INDUCTION = store_thm
  ("NUM_ADMIT_WF_INDUCTION",
   Term `ADMIT_WF_INDUCTION $<`,
    REWRITE_TAC [ADMIT_WF_INDUCTION]
    THEN REPEAT STRIP_TAC
    THEN ASSUME_TAC ((BETA_RULE o SPEC (Term `\y. (!x. x<=y ==> P x)`)) 
                      numTheory.INDUCTION)
    THEN IMP_RES_TAC NUM_ADMIT_lemma1
    THEN IMP_RES_TAC NUM_ADMIT_lemma2
    THEN RES_TAC 
    THEN RULE_ASSUM_TAC (REWRITE_RULE [LESS_OR_EQ])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE[] o SPECL [Term `x:num` , Term `x:num`])
    THEN ASM_REWRITE_TAC[]) ;

(*------------------------------------------------------------------------------
   Define the 'joint' or intersection of two relations.
------------------------------------------------------------------------------*)

val REL_JOINT = Define `REL_JOINT (U:^Rel) V = \x y. U x y /\ V x y` ;
val _ = set_fixity "REL_JOINT" (Infix 600) ;

(*-------------------------------------------------------------------------------
  REL_JOINT_ADMIT_WF_INDUCTION: 
  states that the joint (intersection) of two relations preserves WF_INDUCTION
-------------------------------------------------------------------------------*)


val REL_JOINT_ADMIT_sWF_Ind = store_thm
  ("REL_JOINT_ADMIT_sWF_Ind",
   Term `!A U (V:^Rel). 
      ADMIT_sWF_Ind A U ==> ADMIT_sWF_Ind A  (U REL_JOINT V)`,
    REWRITE_TAC [REL_JOINT, ADMIT_sWF_Ind]
    THEN BETA_TAC THEN REPEAT_FIN 6 STRIP_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC THEN RES_TAC
    THEN SUBGOAL_THEN (Term `(!x. x IN A /\ (U:^Rel) x y /\ V x y ==> P x)`)
         ASSUME_TAC
    THENL [ REPEAT STRIP_TAC THEN RES_TAC , RES_TAC ]) ;

val REL_JOINT_ADMIT_WF_IND = store_thm
  ("REL_JOINT_ADMIT_WF_IND",
   Term `!U (V:^Rel). ADMIT_WF_INDUCTION U ==> ADMIT_WF_INDUCTION (U REL_JOINT V)`,
    REWRITE_TAC [ADMIT_sWF_WF, REL_JOINT_ADMIT_sWF_Ind]) ;

(*------------------------------------------------------------------------------
   Definition of the product of two relations
------------------------------------------------------------------------------*)

val REL_PROD = Define 
    `REL_PROD (U:^Rel1) (V:^Rel2) = 
    (\x y. U (FST x) (FST y) /\ V (SND x) (SND y))` ;
val _ = set_fixity "REL_PROD" (Infix 600) ;

val PROD = Define `PROD (A:('a)set) (B:('b)set) = {(x,y) | x IN A /\ y IN B}` ;
val _ = set_fixity "PROD" (Infix 400) ;

(*-------------------------------------------------------------------------------
  REL_PROD_ADMIT_WF_INDUCTION: 
  states that the product of two relations preserves WF_INDUCTION
-------------------------------------------------------------------------------*)

val REL_PROD_ADMIT_sWF_Ind = store_thm
  ("REL_PROD_ADMIT_sWF_Ind",
   Term `!A B (U:^Rel1) (V:^Rel2). 
         ADMIT_sWF_Ind A U ==> ADMIT_sWF_Ind (A PROD B) (U REL_PROD V)`,
    REWRITE_TAC [PROD, REL_PROD, ADMIT_sWF_Ind]
    THEN BETA_TAC THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC (Term `\x1:'a. (!x2:'b. x2 IN B ==> P (x1,x2))`))
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN SUBGOAL_THEN 
         (Term `!y:'a. y IN A /\ (!x. x IN A /\ U x y ==> (!x2. x2 IN B ==> P(x,x2))) 
                ==>
               (!x2:'b. x2 IN B==> P(y,x2))`)
         ASSUME_TAC
    THENL 
    [ (*-1-*)
      REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC 
      THEN REPEAT STRIP_TAC
      THENL 
      [ EXISTS_TAC (Term `y':'a`) THEN EXISTS_TAC (Term `x2:'b`)
        THEN ASM_REWRITE_TAC[] ,
        ASM_REWRITE_TAC[]
        THEN XRULE_ASSUM_TAC (SUBS [ASSUME (Term `(x'':'a#'b) = (x''',y'')`)])
        THEN XRULE_ASSUM_TAC (REWRITE_RULE [pairTheory.FST, pairTheory.SND])
        THEN RES_TAC ] ,
      (*-2-*)
      ASM_REWRITE_TAC[] THEN RES_TAC ]
   ) ;

val UNIV_prod_UNIV = prove
  (Term `(UNIV:('a)set) PROD (UNIV:('b)set) = UNIV`,
    REWRITE_TAC [PROD, EXTENSION, IN_UNIV]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REWRITE_TAC [IN_UNIV] THEN GEN_TAC
    THEN EXISTS_TAC (Term `FST (x:'a#'b)`)
    THEN EXISTS_TAC (Term `SND (x:'a#'b)`)
    THEN REWRITE_TAC [pairTheory.PAIR]) ;
 
val REL_PROD_ADMIT_WF_IND = store_thm
  ("REL_PROD_ADMIT_WF_IND",
   Term `!(U:^Rel1) (V:^Rel2). 
         ADMIT_WF_INDUCTION U ==> ADMIT_WF_INDUCTION (U REL_PROD V)`,
    REWRITE_TAC [ADMIT_sWF_WF, (SYM UNIV_prod_UNIV), REL_PROD_ADMIT_sWF_Ind]) ;

(*------------------------------------------------------------------------------
  definition of lexicographic product of two relations
------------------------------------------------------------------------------*)

val REL_LEXI = Define
   `REL_LEXI (U:^Rel1) (V:^Rel2) = 
    (\x y. U (FST x) (FST y) \/ (((FST x)=(FST y)) /\ V (SND x) (SND y)))` ;

val _ = set_fixity "REL_LEXI" (Infix 540) ;


val Dpred = ty_antiq (Type `:('a#'b)->bool`) ;
val FIRST  = Term `\(P:^Dpred) x. ?y. y IN B /\ P(x,y)` ;

(*-------------------------------------------------------------------------------
  Now we are going to prove that lexicographic product preserve the WF 
  induction.
-------------------------------------------------------------------------------*)

val LEXI_lemma1 = (BETA_RULE o prove)
  (Term `!(P:^Dpred) (R:^Rel1) X. 
        X IN (A PROD B) /\ P X /\ sWF A R 
         ==>
        (?x2. x2 IN A /\ (^FIRST P)x2 /\ 
              (!x1. x1 IN A /\ (^FIRST P)x1 ==> ~(R x1 x2)))`,
    REWRITE_TAC [PROD, sWF_CONTRA] THEN BETA_TAC
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC (Term `^FIRST P`))
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN XRULE_ASSUM_TAC (SUBS [ASSUME (Term `(X:'a#'b)=(x,y)`)])
    THEN SUBGOAL_THEN (Term `?x':'a. x' IN A /\ (?y':'b. y' IN B /\ P(x',y'))`)
         (XRULE_ASSUM_TAC o (fn th1=>fn th2=> MP th2 th1))
    THENL [ EXISTS_TAC (Term `x:'a`) THEN ASM_REWRITE_TAC[]
            THEN EXISTS_TAC (Term `y:'b`) THEN ASM_REWRITE_TAC[] ,
            ALL_TAC ]
    THEN ASM_REWRITE_TAC[]) ;

val LEXI_lemma2 = (BETA_RULE o prove)
  (Term `!(P:^Dpred) (U:^Rel1) (V:^Rel2) X m.
         m IN A /\ X IN (A PROD B) /\ P X /\
         (!Y. Y IN (A PROD B) /\ P Y  ==> 
              (?X.  X IN (A PROD B) /\ (U REL_LEXI V) X Y  /\  P X))  /\
         (!a. a IN A /\ (^FIRST P) a ==> ~(U a m))
         ==>
         (!v. v IN B /\ P(m,v) ==> (?u. u IN B /\ V u v /\ P(m,u)))`,
    REWRITE_TAC [PROD, REL_LEXI] THEN BETA_TAC
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC (Term `((m:'a),(v:'b))`))
    THEN SUBGOAL_THEN 
         (Term `(?(x:'a) (y:'b). ((m,v)=(x,y)) /\ x IN A /\ y IN B) /\ P(m,v)`)
         (XRULE_ASSUM_TAC o (fn th1=>fn th2=> MP th2 th1))
    THENL
    [ ASM_REWRITE_TAC[] THEN EXISTS_TAC (Term `m:'a`) 
      THEN EXISTS_TAC (Term `v:'b`)
      THEN ASM_REWRITE_TAC[] , ALL_TAC ]
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SUBS [ASSUME (Term `(X':'a#'b) = (x',y')`)])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [pairTheory.FST, pairTheory.SND])
    THENL
    [ (*- 1 -*)
      UNDISCH_TAC (Term `T`) THEN RES_TAC ,
      (*- 2 -*)
      EXISTS_TAC (Term `y':'b`) THEN UNDISCH_TAC (Term `T`)
      THEN XRULE_ASSUM_TAC SYM
      THEN ASM_REWRITE_TAC[]
    ]) ;

val LEXI_lemma3 = prove
  (Term `!(U:^Rel1) (V:^Rel2).
         ~(sWF (A PROD B) (U REL_LEXI V)) /\ sWF A U 
         ==>
         (?P. (?x. x IN B /\  P x) /\ 
              (!v. v IN B /\ P v ==> (?u. u IN B /\ V u v /\ P u)))`,
    REWRITE_TAC [NOT_sWF] 
    THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN
         (Term `(?(x2:'a).
                  x2 IN A /\
                  (?(y:'b). y IN B /\ P (x2,y)) /\
                  (!x1. x1 IN A /\ (?y. y IN B /\ P (x1,y)) ==> ~(U x1 x2)))`)
         ASSUME_TAC
    THENL
    [ MATCH_MP_TAC LEXI_lemma1
      THEN EXISTS_TAC (Term `x:'a#'b`) THEN ASM_REWRITE_TAC[] , 
      ALL_TAC ]
    THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
    THEN EXISTS_TAC (Term `\y. (P:('a#'b)->bool) (x2:'a,y)`) 
    THEN BETA_TAC THEN CONJ_TAC
    THENL
    [ EXISTS_TAC (Term `y:'b`)      
      THEN ASM_REWRITE_TAC [pairTheory.SND] ,
      ALL_TAC ]
    THEN MATCH_MP_TAC LEXI_lemma2 
    THEN EXISTS_TAC (Term `U:^Rel1`)
    THEN EXISTS_TAC (Term `x:'a#'b`)
    THEN ASM_REWRITE_TAC[] 
  ) ;
      
(*---------------------------------------------------------------------------------
  REL_LEXI_ADMIT_WF_IND:
  States that lexicographic product does preserve the WF induction.
---------------------------------------------------------------------------------*)

val REL_LEXI_ADMIT_sWF_Ind = store_thm
  ("REL_LEXI_ADMIT_sWF_Ind",
   Term `!A B (U:^Rel1) (V:^Rel2).
        ADMIT_sWF_Ind A U  /\  ADMIT_sWF_Ind B V
        ==>
        ADMIT_sWF_Ind (A PROD B) (U REL_LEXI V)`,
    REWRITE_TAC [(SYM o SPEC_ALL) sWF_EQU_IND]
    THEN REPEAT STRIP_TAC
    THEN ASM_CASES_TAC (Term `sWF ( A PROD B) ((U:^Rel1) REL_LEXI (V:^Rel2))`)
    THEN ASM_REWRITE_TAC[]
    THEN IMP_RES_TAC LEXI_lemma3 
    THEN XRULE_ASSUM_TAC (SUBS [ISPECL [Term `B:('b)set`, Term `V:^Rel2`] 
                          sWF_CONTRA])
    THEN SUBGOAL_THEN 
         (Term `?y:'b. y IN B /\ P' y /\ (!x. x IN B /\ P' x ==> ~V x y)`)
         (fn thm=> ASSUME_TAC thm THEN UNDISCH_TAC (concl thm))
    THENL [ FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC (Term `x':'b`) 
            THEN ASM_REWRITE_TAC[] , ALL_TAC ]
    THEN REWRITE_TAC [PROVE [] `p==>F=~p`]
    THEN CONV_TAC NOT_EXISTS_CONV
    THEN REWRITE_TAC [PROVE [] `~(p/\q/\r)=p/\q==>~r`]
    THEN (CONV_TAC o DEPTH_CONV) NOT_FORALL_CONV
    THEN ASM_REWRITE_TAC [DECIDE `~(p/\q==>~r)=p/\r/\q`] ) ;


val REL_LEXI_ADMIT_WF_IND = store_thm
  ("REL_LEXI_ADMIT_WF_IND",
   Term `!(U:^Rel1) (V:^Rel2).
        ADMIT_WF_INDUCTION U  /\  ADMIT_WF_INDUCTION V
        ==>
        ADMIT_WF_INDUCTION (U REL_LEXI V)`,
    REWRITE_TAC [ ADMIT_sWF_WF, SYM( SPEC_ALL UNIV_prod_UNIV), 
                  REL_LEXI_ADMIT_sWF_Ind ]) ;

(*---------------------------------------------------------------------------------
  Now we are going to define what is a finite product on relations, then proves
  that this preserves WF induction. 

  Due to the chosen representation, the relations are required to be all on the 
  same type.

  Finitely many relations that we are going to combine will be represented by
  a function R of type num -> Rel and a number n to mark that we are interested in
  relations R 0 ... R n.

  Assuming Rel = 'a->'a->bool, the n-product of R will be a relation over vectors
  where a vector is an object of type num->'a. 
       
---------------------------------------------------------------------------------*)

val Rel_Vec = ty_antiq (Type `:num->^Rel`);
val Vector = ty_antiq (Type `:num->'a`) ;

val REL_FPROD = Define
   `(REL_FPROD 0 (R:^Rel_Vec) (s:^Vector) (t:^Vector) = (R 0) (s 0) (t 0)) /\
    (REL_FPROD (SUC n) R s t = 
       REL_FPROD n R s t /\ (R (SUC n)) (s (SUC n)) (t (SUC n)))` ;

val REL_FPROD_SUC = store_thm
  ("REL_FPROD_SUC",
   Term `!(R:^Rel_Vec) s t n.
       REL_FPROD (SUC n) R s t = 
       ((REL_FPROD n R) REL_PROD (R (SUC n))) (s,s(SUC n)) (t,t(SUC n))`,
    REWRITE_TAC [REL_FPROD, REL_PROD] THEN BETA_TAC
    THEN REWRITE_TAC [pairTheory.FST, pairTheory.SND]) ;

val leq0_0 = DECIDE `(x<=0) = (x=0)` ;

val eq0 = PROVE [] `(!i. (i=0) ==> P i) = P 0` ;

fun eq0_CONV trm =
    let 
      val i = #Bvar (dest_forall trm)
      val P' = (#conseq o dest_imp o #Body o dest_forall) trm 
      val P = mk_abs {Bvar=i, Body=P'} 
      val left_imp = (#ant o dest_imp o #Body o dest_forall) trm 
      val body = mk_imp {ant=left_imp, conseq=mk_comb{Rand=i,Rator=P}} 
    in
    (BETA_RULE o prove)
    ( mk_eq{lhs = mk_forall{Bvar=i,Body=body}, 
            rhs = mk_comb{Rand=(Term`0`), Rator=P}},
      REWRITE_TAC[eq0]) 
    end ;

val LESS_EQ_SUC = DECIDE `m <= (SUC n) = (m <= n) \/ (m = SUC n)` ;

val REL_FPROD_ALL = store_thm
  ("REL_FPROD_ALL",
   Term `!(R:^Rel_Vec) s t n.
       REL_FPROD n R s t = (!i. i<=n ==> (R i) (s i) (t i))`,
   REPEAT_FIN 3 GEN_TAC
   THEN Induct_on `n`
   THENL
   [ (*- 1 -*)
     REWRITE_TAC [REL_FPROD, leq0_0]
     THEN PROVE_TAC[] ,
     (*- 2 -*)
     ASM_REWRITE_TAC [REL_FPROD, LESS_EQ_SUC]
     THEN EQ_TAC THEN REPEAT STRIP_TAC
     THENL [ RES_TAC , ASM_REWRITE_TAC[] ,
             RES_TAC , 
             FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[] ]
   ]) ;

(*---------------------------------------------------------------------------------
  FPROD_ADMIT_WF_IND
  States that finite product of N relations preserves WF induction.
---------------------------------------------------------------------------------*)

fun IMP_SPEC trm thm =
    let val asm = (#ant o dest_imp o concl) thm 
    in
    (DISCH asm (SPEC trm (UNDISCH thm))) 
    end ; 

val FPROD_ADMIT_sWF_Ind = store_thm
 ("FPROD_ADMIT_sWF_Ind",
   Term `!A (R:^Rel_Vec) m n.
       m<=n /\ ADMIT_sWF_Ind (A m) (R m)
       ==>
       ADMIT_sWF_Ind {s | !i. i<=n ==> (s i) IN (A i)} (REL_FPROD n R)`,
    REWRITE_TAC [ADMIT_sWF_Ind, REL_FPROD_ALL]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC 
         (SPEC (Term `\x:'a. !s. ((s m) = x) /\ (!i. i<=n ==> (s i) IN (A i)) ==> P s`))
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN XRULE_ASSUM_TAC (IMP_SPEC (Term `(x:^Vector) m`))
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [PROVE [] `p==>q==>r=p/\q==>r`])
    THEN XRULE_ASSUM_TAC (IMP_SPEC (Term `x:^Vector`))
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [])
    THEN SUBGOAL_THEN (Term `((x:^Vector) m) IN (A m)`) ASSUME_TAC
    THENL [ RES_TAC , ALL_TAC ]
    THEN UNDISCH_TAC (Term `!i. i <= n ==> ((x:^Vector) i) IN (A i)`)
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT STRIP_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC
    THENL 
    [ RES_TAC ,
      XRULE_ASSUM_TAC (SUBS [SYM (ASSUME (Term `(s:^Vector) m = y`))])
      THEN XRULE_ASSUM_TAC (SPEC (Term `(x':^Vector) m`))
      THEN XRULE_ASSUM_TAC (IMP_SPEC (Term `x':^Vector`))
      THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
      THEN REPEAT RES_TAC ]
   ) ;

val vector_UNIV = prove
  (Term `{s:^Vector | !i. i<=n ==> (s i) IN (UNIV)} = UNIV`,
    REWRITE_TAC [EXTENSION]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REWRITE_TAC [IN_UNIV]) ;

val lemma = 
    REWRITE_RULE [vector_UNIV]
    (BETA_RULE (SPEC (Term `\m:num. UNIV:('a)set`) FPROD_ADMIT_sWF_Ind)) ;

val FPROD_ADMIT_WF_IND = store_thm
 ("FPROD_ADMIT_WF_IND",
   Term `!(R:^Rel_Vec) m n.
       m<=n /\ ADMIT_WF_INDUCTION (R m)
       ==>
       ADMIT_WF_INDUCTION (REL_FPROD n R)`,
    REWRITE_TAC [ADMIT_sWF_WF, lemma]) ;

(*---------------------------------------------------------------------------------
  Now we are going to define what is a finite lexicographic product of relations, 
  then proves that this preserves WF induction. 

  Due to the chosen representation, the relations are required to be all on the 
  same type.
---------------------------------------------------------------------------------*)

val Tail = Define `Tail (s:^Vector) = (\i. s (SUC i))` ;

val Cons = Define `(Cons (a:'a) s 0 = a) /\ (Cons a s (SUC i) = s i)` ;

val REL_FLEXI = Define
   `(REL_FLEXI 0 (R:^Rel_Vec) (s:^Vector) (t:^Vector) = (R 0) (s 0) (t 0)) /\
    (REL_FLEXI (SUC n) R s t = 
       (R 0) (s 0) (t 0)  \/
       ((s 0 = t 0) /\ REL_FLEXI n (Tail R) (Tail s) (Tail t)))` ;

val REL_FLEXI_SUC = store_thm
  ("REL_FLEXI_SUC",
   Term `!(R:^Rel_Vec) s t n. 
        REL_FLEXI (SUC n) R s t =
        ((R 0) REL_LEXI (REL_FLEXI n (Tail R))) (s 0, Tail s) (t 0, Tail t)`,
    REWRITE_TAC [REL_FLEXI, REL_LEXI] THEN BETA_TAC
    THEN REWRITE_TAC [pairTheory.FST, pairTheory.SND]) ;

val REL_FLEXI_ALL = store_thm
  ("REL_FLEXI_ALL",
   Term `!n (R:^Rel_Vec) s t.
         REL_FLEXI n R s t = 
         (?m. m<=n /\ (!i. i<m ==> (s i = t i)) /\ (R m) (s m) (t m))`,
    Induct_on `n`
    THENL
    [ (*- 1 -*)
      REPEAT GEN_TAC THEN REWRITE_TAC [REL_FLEXI, leq0_0] 
      THEN EQ_TAC THEN REPEAT STRIP_TAC
      THENL
      [ EXISTS_TAC (Term `0`) THEN ASM_REWRITE_TAC [prim_recTheory.NOT_LESS_0] ,
        FIRST_ASSUM (SUBST1_TAC o SYM) THEN ASM_REWRITE_TAC [] ] ,
      (*- 2 -*)
      REPEAT GEN_TAC THEN ASM_REWRITE_TAC [REL_FLEXI, Tail]
      THEN BETA_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
      THENL 
      [ (*- 2.1 -*)
        EXISTS_TAC (Term `0`) 
        THEN ASM_REWRITE_TAC [prim_recTheory.NOT_LESS_0, ZERO_LESS_EQ] ,
        (*- 2.2 -*)
        EXISTS_TAC (Term `SUC m`) THEN REPEAT STRIP_TAC
        THENL 
        [ ASM_REWRITE_TAC [LESS_EQ_MONO] ,
          UNDISCH_TAC (Term `i < (SUC m)`)
          THEN SPEC_TAC (Term `i:num`,Term `i:num`)
          THEN Induct_on `i`
          THEN ASM_REWRITE_TAC[LESS_MONO_EQ] ,
          ASM_REWRITE_TAC [] ] ,
        (*- 2.3 -*)
        DISJ_CASES_TAC (SPEC (Term `m:num`) num_CASES)
        THENL [ FIRST_ASSUM (SUBST1_TAC o SYM) THEN ASM_REWRITE_TAC [] , 
                UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC ]
        THEN XRULE_ASSUM_TAC (REWRITE_RULE [ ASSUME (Term `m = SUC n'`)])
        THEN UNDISCH_TAC (Term `T`) THEN REWRITE_TAC[]
        THEN DISJ2_TAC THEN CONJ_TAC
        THENL [ FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC [prim_recTheory.LESS_0] ,
                ALL_TAC ]
        THEN EXISTS_TAC (Term `n':num`) THEN REPEAT STRIP_TAC
        THENL [ IMP_RES_TAC LESS_EQ_MONO ,
                FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC [LESS_MONO_EQ] ,
                ASM_REWRITE_TAC [] ]
      ]
    ]) ;
   
val lemma = PROVE [] `((a:'a) IN A ==> p) = (a IN A ==> (a=a) ==> p)` ;

val FLEXI_0 = prove
  (Term `!A (R:^Rel_Vec).
       ADMIT_sWF_Ind A (R 0)  ==> 
       ADMIT_sWF_Ind {s | s 0 IN A} (REL_FLEXI 0 R)`,
    REWRITE_TAC [ADMIT_sWF_Ind, REL_FLEXI]
    THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC (Term `\z:'a. !x. (z = (x 0)) ==> P x`))
    THEN XRULE_ASSUM_TAC BETA_RULE 
    THEN UNDISCH_TAC (Term `((x:num->'a) 0) IN A`)
    THEN ONCE_REWRITE_TAC [lemma]
    THEN (XRULE_ASSUM_TAC o CONV_RULE o DEPTH_CONV) RIGHT_IMP_FORALL_CONV
    THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
    THEN REPEAT STRIP_TAC 
    THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
    THEN FIRST_ASSUM (fn thm=> SUBST1_TAC (SYM thm) THEN DEL_TAC (concl thm))
    THEN ASM_REWRITE_TAC[]
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPECL [Term `(x':num->'a) 0`, Term `x':num->'a`])
    THEN XRULE_ASSUM_TAC (REWRITE_RULE[])
    THEN RES_TAC ) ;

val lemma1 = prove
  (Term `(?s:^Vector. (s 0 = x 0) /\ (Tail s = Tail x) /\ P s) ==> P x`,
    REWRITE_TAC [Tail] 
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN (Term `(x:^Vector)=s`) (fn thm=> ASM_REWRITE_TAC [thm])
    THEN CONV_TAC FUN_EQ_CONV
    THEN Induct_on `n` THEN ASM_REWRITE_TAC[]
    THEN (XRULE_ASSUM_TAC o CONV_RULE) FUN_EQ_CONV
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN ASM_REWRITE_TAC[]) ;

val lemma2 = prove
  (Term `((s:^Vector) 0 = x 0) /\ (Tail s = Tail x) /\ P s ==> P x`,
    REWRITE_TAC [Tail] 
    THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN (Term `(x:^Vector)=s`) (fn thm=> ASM_REWRITE_TAC [thm])
    THEN CONV_TAC FUN_EQ_CONV
    THEN Induct_on `n` THEN ASM_REWRITE_TAC[]
    THEN (XRULE_ASSUM_TAC o CONV_RULE) FUN_EQ_CONV
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN ASM_REWRITE_TAC[]) ;

val lemma3 = prove 
  (Term `(Cons(FST y)(SND y)0,Tail(Cons(FST y)(SND y))) = (y:'a#^Vector)`,
    REWRITE_TAC [Tail, Cons]
    THEN (CONV_TAC o DEPTH_CONV) ETA_CONV
    THEN REWRITE_TAC [pairTheory.PAIR]) ;

val LEXI_FLEXI = prove
  (Term `!(U:^Rel_Vec). 
       ADMIT_sWF_Ind ((A 0) PROD {s| !i. i<=n ==> (s i) IN ((Tail A) i)})
                     ((U 0) REL_LEXI (REL_FLEXI n (Tail U)))
       ==>
       ADMIT_sWF_Ind {s| !i. i<=(SUC n) ==> (s i) IN (A i)}
                     (REL_FLEXI (SUC n) U)`,
    REWRITE_TAC [ADMIT_sWF_Ind, REL_FLEXI_SUC, PROD]
    THEN REPEAT STRIP_TAC
    THEN XRULE_ASSUM_TAC (SPEC 
         (Term `\X:'a#^Vector. ?s. (s 0 = FST X) /\ (Tail s = SND X) /\ P s`))
    THEN XRULE_ASSUM_TAC BETA_RULE
    THEN XRULE_ASSUM_TAC (IMP_SPEC (Term `(x:^Vector) 0, Tail x`))
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [pairTheory.FST, pairTheory.SND])
    THEN MATCH_MP_TAC lemma1
    THEN XRULE_ASSUM_TAC (REWRITE_RULE [PROVE [] `p==>q==>r=p/\q==>r`])
    THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
    THEN REPEAT STRIP_TAC
    THENL
    [ (*-1-*)
      EXISTS_TAC (Term `Cons ((FST y):'a) (SND y)`)
      THEN REWRITE_TAC [Tail, Cons] THEN BETA_TAC
      THEN (CONV_TAC o DEPTH_CONV) ETA_CONV THEN REWRITE_TAC []
      THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
      THEN REPEAT STRIP_TAC
      THENL
      [ (*-1.1-*)
        (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
        THEN Induct_on `i` THEN REWRITE_TAC [ZERO_LESS_EQ, Cons]
        THENL
        [ (XRULE_ASSUM_TAC o CONV_RULE o DEPTH_CONV) SET_SPEC_CONV
          THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
          THEN ASM_REWRITE_TAC [pairTheory.FST] ,
          (XRULE_ASSUM_TAC o CONV_RULE o DEPTH_CONV) SET_SPEC_CONV
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [Tail])
          THEN XRULE_ASSUM_TAC BETA_RULE
          THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
          THEN ASM_REWRITE_TAC [pairTheory.SND] 
          THEN FIRST_ASSUM MATCH_MP_TAC
          THEN IMP_RES_TAC LESS_EQ_MONO ] ,
        (*-1.2-*)
        (XRULE_ASSUM_TAC o CONV_RULE o DEPTH_CONV) RIGHT_IMP_EXISTS_CONV
         THEN XRULE_ASSUM_TAC (SPEC (Term `(x':^Vector)0, Tail x'`))
         THEN UNDISCH_ALL_TAC
         THEN REWRITE_TAC [pairTheory.FST, pairTheory.SND] THEN REPEAT STRIP_TAC
         THEN MATCH_MP_TAC lemma2
         THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
         THEN XRULE_ASSUM_TAC (REWRITE_RULE [lemma3])
         THEN ASM_REWRITE_TAC []
         THEN CONV_TAC SET_SPEC_CONV
         THEN EXISTS_TAC (Term `(x':^Vector) 0`) 
         THEN EXISTS_TAC (Term `Tail (x':^Vector)`)
         THEN REWRITE_TAC [Tail] 
         THEN UNDISCH_TAC 
              (Term `x' IN {s:^Vector | !i. i <= (SUC n) ==> (s i) IN (A i)}`)
         THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
         THEN BETA_TAC THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC 
         THENL [ REWRITE_TAC [ZERO_LESS_EQ] , ASM_REWRITE_TAC [LESS_EQ_MONO]] 
       ] ,
     (*-2-*)
     (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
     THEN EXISTS_TAC (Term `(x:^Vector) 0`) 
     THEN EXISTS_TAC (Term `Tail (x:^Vector)`)
     THEN REWRITE_TAC [Tail] 
     THEN UNDISCH_TAC (Term `x IN {s:^Vector | !i. i <= (SUC n) ==> (s i) IN (A i)}`)
     THEN (CONV_TAC o DEPTH_CONV) SET_SPEC_CONV
     THEN BETA_TAC THEN REPEAT STRIP_TAC
     THEN FIRST_ASSUM MATCH_MP_TAC 
     THENL [ REWRITE_TAC [ZERO_LESS_EQ] , ASM_REWRITE_TAC [LESS_EQ_MONO]]
   ]) ;


val FLEXI_lemma = prove
  (Term `!(U:^Rel_Vec). 
       ADMIT_sWF_Ind (A 0) (U 0) /\
       ADMIT_sWF_Ind {s | !i. i <= n ==> (s i) IN (Tail A i)}
                     (REL_FLEXI n (Tail U))
       ==>
       ADMIT_sWF_Ind {s | !i. i <= (SUC n) ==> (s i) IN (A i)}
                     (REL_FLEXI (SUC n) U)`,
    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC LEXI_FLEXI
    THEN IMP_RES_TAC REL_LEXI_ADMIT_sWF_Ind) ;

(*---------------------------------------------------------------------------------
  FLEXI_ADMIT_WF_IND
  States that finite lexicographic product of N relations preserves WF induction.
---------------------------------------------------------------------------------*)

val SUC_MONO = DECIDE `!n m. (n<=m) ==> ((SUC n)<=(SUC m))` ;

val FLEXI_ADMIT_sWF_Ind = store_thm
  ("FLEXI_ADMIT_sWF_Ind",
   Term `!A (R:^Rel_Vec) n.
        (!i. (i<=n) ==> ADMIT_sWF_Ind (A i) (R i)) 
         ==>
        ADMIT_sWF_Ind {s | !i. i <= n ==> (s i) IN (A i)} (REL_FLEXI n R)`,
    REPEAT GEN_TAC
    THEN SPEC_TAC (Term `R:^Rel_Vec`, Term `R:^Rel_Vec`)
    THEN SPEC_TAC (Term `A:num->(('a)set)`, Term `A:num->(('a)set)`)
    THEN SPEC_TAC (Term `n:num`, Term `n:num`)
    THEN Induct_on `n`
    THENL
    [ REWRITE_TAC [leq0_0] 
      THEN (CONV_TAC o DEPTH_CONV) eq0_CONV 
      THEN REWRITE_TAC [FLEXI_0] ,
      REPEAT STRIP_TAC 
      THEN MATCH_MP_TAC FLEXI_lemma THEN CONJ_TAC
      THENL
      [ FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC [ZERO_LESS_EQ] ,
        XRULE_ASSUM_TAC 
              (SPECL [Term `Tail (A:num->(('a)set))`, Term `Tail (R:^Rel_Vec)`])
        THEN FIRST_ASSUM (fn thm=> MATCH_MP_TAC thm THEN DEL_TAC (concl thm))
        THEN REWRITE_TAC [Tail] THEN BETA_TAC
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC [LESS_EQ_MONO]
      ]
    ]) ;

val lemma = 
    REWRITE_RULE [vector_UNIV]
    (BETA_RULE (SPEC (Term `\m:num. UNIV:('a)set`) FLEXI_ADMIT_sWF_Ind)) ;

val FLEXI_ADMIT_WF_IND  = store_thm
  ("FLEXI_ADMIT_WF_IND",
   Term `!(R:^Rel_Vec) n.
        (!i. (i<=n) ==> ADMIT_WF_INDUCTION (R i))
         ==>
        ADMIT_WF_INDUCTION (REL_FLEXI n R)`,
    REWRITE_TAC [ADMIT_sWF_WF, lemma]) ;

val _ = export_theory() ;

(*--- DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG DCG ---*)
